# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2022 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <https://www.gnu.org/licenses/>.
#
import gettext
import copy
import decimal
import fractions
from typing import Optional, Union, Dict, List, Tuple, cast

from PyQt6.QtWidgets import (
    QFormLayout,
    QLabel,
    QMessageBox,
    QPushButton,
    QVBoxLayout,
    QWidget,
    QDialog,
    QDialogButtonBox,
)

from PyQt6.QtCore import Qt, pyqtBoundSignal, QAbstractTableModel, QModelIndex, QVariant

from . import widgets
from . import dialogs
from .. import utils
from .. import scoring
from .. import exams

t = gettext.translation("eyegrade", utils.locale_dir(), fallback=True)
_ = t.gettext


class ScoreWeightsTableModel(QAbstractTableModel):
    """Table for editing score weight values."""

    changes: bool
    exam_config: exams.ExamConfig
    models: List[str]
    has_permutations: bool
    permutations: Dict[str, List[Tuple[int, List[int]]]]
    weights: List[List[Union[int, float, decimal.Decimal, fractions.Fraction]]]
    sum_weights: List[Union[int, float, decimal.Decimal, fractions.Fraction]]

    def __init__(self, exam_config: exams.ExamConfig, parent=None):
        super().__init__(parent=parent)
        self.exam_config = exam_config
        self.data_reset()
        self.changes = False
        cast(pyqtBoundSignal, self.dataChanged).connect(self._update_weights_sum)

    def data_reset(self) -> None:
        self.beginResetModel()
        self.models = sorted(self.exam_config.models)
        self.has_permutations = False
        if not self.models:
            self.models = ["A"]
        elif all(self.exam_config.get_permutations(m) for m in self.models):
            self.has_permutations = True
            self.models.insert(0, "0")
            self.permutations = self.exam_config.permutations
        if not self.has_permutations:
            self.weights = [
                self.exam_config.get_question_weights(m) for m in self.models
            ]
            for i in range(len(self.weights)):
                if not self.weights[i]:
                    self.weights[i] = [1] * self.exam_config.num_questions
        else:
            weights_0: list[Union[int, float, decimal.Decimal, fractions.Fraction]]
            weights_0 = [1] * self.exam_config.num_questions
            weights_m = self.exam_config.get_question_weights(self.models[1])
            if weights_m:
                for i, value in enumerate(weights_m):
                    weights_0[self._to_model_0(1, i)] = value
            else:
                weights_0 = [1] * self.exam_config.num_questions
            self.weights = [weights_0]
        self.sum_weights = [sum(w) for w in self.weights]
        self.changes = False
        self.endResetModel()

    def clear(self) -> None:
        self.beginResetModel()
        for i in range(len(self.weights)):
            self.weights[i] = [1] * self.exam_config.num_questions
        self.sum_weights = [sum(w) for w in self.weights]
        self.changes = False
        self.endResetModel()

    def validate(self) -> bool:
        """Checks that the weights are valid.

        When there are no permutations, it checks that all the models
        have the same weight elements, regardless their order.

        """
        if self.has_permutations:
            # The weights are necessarily valid
            valid = True
        else:
            # All the weights must have the same elements
            reference = sorted(self.weights[0])
            for w in self.weights[1:]:
                if reference != sorted(w):
                    valid = False
                    break
            else:
                valid = True
        return valid

    def consolidate(self) -> bool:
        """Consolidate the weights into the exam_config object.

        Returns True if the values are correct and consolidated,
        and False if not.

        """
        if not self.validate():
            return False
        self.exam_config.reset_question_weights()
        if self.has_permutations:
            for i, model in enumerate(self.models[1:]):
                w = []
                model_idx = i + 1
                for q in range(self.exam_config.num_questions):
                    w.append(self.weights[0][self._to_model_0(model_idx, q)])
                self.exam_config.set_question_weights(model, w)
        else:
            # All the weights must have the same elements
            for model, w in zip(self.models, self.weights):
                self.exam_config.set_question_weights(model, w)
        return True

    def rowCount(self, parent=QModelIndex()) -> int:
        return self.exam_config.num_questions + 1

    def columnCount(self, parent=QModelIndex()) -> int:
        return len(self.models)

    def headerData(
        self,
        section: int,
        orientation: Qt.Orientation,
        role: int = Qt.ItemDataRole.DisplayRole,
    ) -> Union[str, QVariant]:
        if role == Qt.ItemDataRole.DisplayRole:
            if orientation == Qt.Orientation.Horizontal:
                return "{0} {1}".format(_("Model"), self.models[section])
            else:
                if section < self.exam_config.num_questions:
                    return "{0} {1}".format(_("Question"), section + 1)
                else:
                    return _("Total")
        else:
            return QVariant()

    def data(self, index: QModelIndex, role: int = Qt.ItemDataRole.DisplayRole):
        if role == Qt.ItemDataRole.DisplayRole:
            r = index.row()
            c = index.column()
            if r < self.exam_config.num_questions:
                if not self.has_permutations or c == 0:
                    value = self.weights[c][r]
                else:
                    value = self.weights[0][self._to_model_0(c, r)]
                return scoring.format_number(value, short=True)
            else:
                if not self.has_permutations:
                    value = self.sum_weights[c]
                else:
                    value = self.sum_weights[0]
                return scoring.format_number(value, short=True, no_fraction=True)
        elif role == Qt.ItemDataRole.TextAlignmentRole:
            return Qt.AlignmentFlag.AlignCenter
        else:
            return QVariant()

    def flags(self, index: QModelIndex) -> Qt.ItemFlag:
        if index.row() < self.exam_config.num_questions:
            return Qt.ItemFlag(Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsEditable)
        else:
            return Qt.ItemFlag(Qt.ItemFlag.ItemIsEnabled)

    def setData(
        self, index: QModelIndex, value_qvar: str, role: int = Qt.ItemDataRole.EditRole
    ) -> bool:
        success = False
        r = index.row()
        c = index.column()
        try:
            value = scoring.parse_number(value_qvar)
        except ValueError:
            return False
        if r < self.exam_config.num_questions:
            if not self.has_permutations or c == 0:
                old_value = self.weights[c][r]
                self.weights[c][r] = value
                r_0 = r
            else:
                r_0 = self._to_model_0(c, r)
                old_value = self.weights[0][r_0]
                self.weights[0][r_0] = value
            if old_value != value:
                self.changes = True
                # Emit the signal for all the affected cells
                data_changed_signal = cast(pyqtBoundSignal, self.dataChanged)
                if not self.has_permutations:
                    data_changed_signal.emit(index, index)
                else:
                    changed_idx = self.index(r_0, 0)
                    data_changed_signal.emit(changed_idx, changed_idx)
                    for model_idx in range(1, len(self.models)):
                        for i in range(self.exam_config.num_questions):
                            if self._to_model_0(model_idx, i) == r_0:
                                changed_idx = self.index(i, model_idx)
                                data_changed_signal.emit(changed_idx, changed_idx)
            success = True
        return success

    def _update_weights_sum(self, index_1: QModelIndex, index_2: QModelIndex) -> None:
        """Handler for the dataChanged signal."""
        if index_1.row() < self.exam_config.num_questions:
            col_1 = index_1.column()
            col_2 = index_2.column()
            if self.has_permutations:
                # Only changes to column 0 are relevant
                col_end = 0
            else:
                col_end = col_2
            for i in range(col_1, col_end + 1):
                self.sum_weights[i] = sum(self.weights[i])
            if col_1 <= col_end:
                if self.has_permutations:
                    # The change affects all the sums
                    col_2 = len(self.models) - 1
                changed_1 = self.index(self.exam_config.num_questions, col_1)
                changed_2 = self.index(self.exam_config.num_questions, col_2)
                data_changed_signal = cast(pyqtBoundSignal, self.dataChanged)
                data_changed_signal.emit(changed_1, changed_2)

    def _to_model_0(self, model_idx: int, question: int) -> int:
        question_0, opt = self.permutations[self.models[model_idx]][question]
        return question_0 - 1


class DialogEditScores(QDialog):
    """Dialog to edit exam scores"""

    exam_config: exams.ExamConfig
    scores_widget: "EditScoresWidget"

    def __init__(self, exam_config: exams.ExamConfig, parent=None):
        super().__init__(parent=parent)
        self.setWindowTitle(_("Edit question scores"))
        layout = QVBoxLayout(self)
        self.setLayout(layout)
        self.scores_widget = EditScoresWidget(exam_config)
        layout.addWidget(self.scores_widget)
        buttons = QDialogButtonBox(
            (
                QDialogButtonBox.StandardButton.Ok
                | QDialogButtonBox.StandardButton.Cancel
            )
        )
        layout.addWidget(buttons)
        accepted_signal = cast(pyqtBoundSignal, buttons.accepted)
        rejected_signal = cast(pyqtBoundSignal, buttons.rejected)
        accepted_signal.connect(self.accept)
        rejected_signal.connect(self.reject)

    def exec(self) -> bool:
        """Shows the dialog and waits until it is closed.

        Returns true if the configuration has been modified,
        or false otherwise.

        """
        terminate = False
        changes = False
        while not terminate:
            result = super().exec()
            if result == QDialog.DialogCode.Accepted:
                old_scores_mode = self.scores_widget.exam_config.scores_mode
                old_base_scores = copy.deepcopy(
                    self.scores_widget.exam_config.base_scores
                )
                old_scores = copy.deepcopy(self.scores_widget.exam_config.scores)
                if self.scores_widget.validate_and_consolidate():
                    terminate = True
                    if (
                        old_scores_mode != self.scores_widget.exam_config.scores_mode
                        or old_base_scores != self.scores_widget.exam_config.base_scores
                        or old_scores != self.scores_widget.exam_config.scores
                    ):
                        changes = True
            else:
                terminate = True
        return changes


class EditScoresWidget(QWidget):
    """Widget for editing score mode, scores and score weights."""

    exam_config: exams.ExamConfig
    current_mode: Optional[int]

    def __init__(self, exam_config: exams.ExamConfig, parent=None):
        super().__init__(parent)
        self.exam_config = exam_config
        self.current_mode = None
        form_widget = QWidget(parent=self)
        table_widget = QWidget(parent=self)
        main_layout = QVBoxLayout(self)
        form_layout = QFormLayout(form_widget)
        table_layout = QVBoxLayout(table_widget)
        main_layout.addWidget(form_widget)
        main_layout.addWidget(table_widget)
        main_layout.setAlignment(table_widget, Qt.AlignmentFlag.AlignHCenter)
        self.combo = widgets.CustomComboBox(parent=self)
        self.combo.set_items(
            [
                _("No scores"),
                _("Same score for all the questions"),
                _("Base score plus per-question weight"),
            ]
        )
        index_changed_signal = cast(pyqtBoundSignal, self.combo.currentIndexChanged)
        index_changed_signal.connect(self._update_combo)
        self.correct_score = widgets.InputScore(is_positive=True)
        correct_score_label = QLabel(_("Score for correct answers"))
        incorrect_score_label = QLabel(_("Score for incorrect answers"))
        blank_score_label = QLabel(_("Score for blank answers"))
        self.incorrect_score = widgets.InputScore(is_positive=False)
        self.blank_score = widgets.InputScore(is_positive=False)
        self.button_reset = QPushButton(_("Reset question weights"))
        button_defaults = QPushButton(_("Compute default scores"))
        self.weights_table = widgets.CustomTableView()
        weights_table_label = QLabel(_("Per-question score weights:"))
        form_layout.addRow(self.combo)
        form_layout.addRow(correct_score_label, self.correct_score)
        form_layout.addRow(incorrect_score_label, self.incorrect_score)
        form_layout.addRow(blank_score_label, self.blank_score)
        table_layout.addWidget(weights_table_label)
        table_layout.addWidget(self.weights_table)
        table_layout.addWidget(self.button_reset)
        table_layout.addWidget(button_defaults)
        table_layout.setAlignment(weights_table_label, Qt.AlignmentFlag.AlignHCenter)
        table_layout.setAlignment(self.weights_table, Qt.AlignmentFlag.AlignHCenter)
        table_layout.setAlignment(self.button_reset, Qt.AlignmentFlag.AlignHCenter)
        table_layout.setAlignment(button_defaults, Qt.AlignmentFlag.AlignHCenter)
        button_reset_clicked_signal = cast(pyqtBoundSignal, self.button_reset.clicked)
        button_reset_clicked_signal.connect(self._reset_weights)
        button_defaults_clicked_signal = cast(pyqtBoundSignal, button_defaults.clicked)
        button_defaults_clicked_signal.connect(self._compute_default_values)
        self.base_score_widgets = [
            self.correct_score,
            correct_score_label,
            self.incorrect_score,
            incorrect_score_label,
            self.blank_score,
            blank_score_label,
            button_defaults,
        ]
        self.weights_widgets = [self.weights_table, weights_table_label]
        self.initialize()

    def initialize(self) -> None:
        model = ScoreWeightsTableModel(self.exam_config, parent=self)
        self.weights_table.setModel(model)
        data_changed_signal = cast(pyqtBoundSignal, model.dataChanged)
        data_changed_signal.connect(self._weights_changed)
        self.weights_table.adjust_size()
        scores = self.exam_config.base_scores
        if scores is not None:
            self._set_score_fields(scores)
        # If the exam is a survey, disable all the controls
        if self.exam_config.survey_mode:
            self.combo.set_item_enabled(1, False)
            self.combo.set_item_enabled(2, False)
            initial_mode = 0
            model.clear()
        else:
            self.combo.set_item_enabled(1, True)
            self.combo.set_item_enabled(2, True)
            if not self.exam_config.scores or self.exam_config.all_weights_are_one():
                model.clear()
            if self.exam_config.scores_mode == exams.ExamConfig.SCORES_MODE_NONE:
                initial_mode = 0
            elif self.exam_config.scores_mode == exams.ExamConfig.SCORES_MODE_WEIGHTS:
                if self.exam_config.all_weights_are_one():
                    initial_mode = 1
                else:
                    initial_mode = 2
            elif (
                self.exam_config.scores_mode == exams.ExamConfig.SCORES_MODE_INDIVIDUAL
            ):
                # This mode cannot be edited in this dialog
                initial_mode = 0
        self.button_reset.setEnabled(False)
        if initial_mode == 0:
            self._enable_weights_widgets(False, False)
        elif initial_mode == 1:
            self._enable_weights_widgets(True, False)
        else:
            self._enable_weights_widgets(True, True)
        self.current_mode = initial_mode
        self.combo.setCurrentIndex(initial_mode)

    def validate_and_consolidate(self) -> bool:
        """Call it to check the values of this page.

        Checks the values and consolidates them into the exam_config object if valid.

        """
        if self.current_mode == 0:
            valid = self._consolidate_no_scores()
        elif self._consolidate_base_scores():
            if self.current_mode == 1:
                valid = True
            else:
                valid = self._consolidate_weights()
        else:
            valid = False
        return valid

    def clear_base_scores(self) -> None:
        self.correct_score.setText("")
        self.incorrect_score.setText("")
        self.blank_score.setText("")

    def _reset_weights(self) -> None:
        model = cast(ScoreWeightsTableModel, self.weights_table.model())
        if model.changes and self._show_warning_weights_reset():
            model.data_reset()

    def _weights_changed(self, index_1, index_2) -> None:
        model = cast(ScoreWeightsTableModel, self.weights_table.model())
        self.button_reset.setEnabled(model.changes)

    def _compute_default_values(self) -> None:
        model = cast(ScoreWeightsTableModel, self.weights_table.model())
        if self.current_mode == 2 and not model.validate():
            self._show_error_weights()
            return
        dialog = dialogs.DialogComputeScores(parent=self)
        score, penalize = dialog.exec()
        if score is None:
            return
        choices = self.exam_config.get_num_choices()
        if self.current_mode == 1:
            # All the questions have the same score
            total_weight = self.exam_config.num_questions
        elif self.current_mode == 2:
            # Weighted questions
            total_weight = model.sum_weights[0]
        else:
            raise NotImplementedError("Bad mode in scores wizard page")
        if self.exam_config.num_questions and choices and choices > 1:
            c_score = score / total_weight
            if penalize:
                i_score = score / (choices - 1) / total_weight
            else:
                i_score = 0
            b_score = 0
            scores = scoring.QuestionScores(c_score, i_score, b_score)
            self._set_score_fields(scores)
        else:
            QMessageBox.critical(
                self,
                _("Error"),
                _("Automatic scores cannot be computed for this exam."),
            )

    def _show_warning_weights_reset(self) -> bool:
        result = QMessageBox.warning(
            self,
            _("Warning"),
            _(
                "The changes you have done to the weights "
                "table will be lost. "
                "Are you sure you want to continue?"
            ),
            QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
            QMessageBox.StandardButton.No,
        )
        return result == QMessageBox.StandardButton.Yes

    def _show_error_weights(self) -> None:
        QMessageBox.critical(
            self,
            _("Error"),
            _(
                "The weights must be the same "
                "in all the models, although they may "
                "be in a different order. "
                "You must fix this before computing "
                "default scores."
            ),
        )

    def _set_score_fields(self, scores: scoring.QuestionScores) -> None:
        self.correct_score.setText(scores.format_correct_score(signed=False))
        self.incorrect_score.setText(scores.format_incorrect_score(signed=True))
        self.blank_score.setText(scores.format_blank_score(signed=True))

    def _update_combo(self, new_index: int) -> None:
        if new_index != self.current_mode:
            # Ask the user if changes to weights may be lost
            model = cast(ScoreWeightsTableModel, self.weights_table.model())
            if (
                self.current_mode == 2
                and model.changes
                and not self._show_warning_weights_reset()
            ):
                self.combo.setCurrentIndex(self.current_mode)
                return
            self.button_reset.setEnabled(False)
            if new_index == 0:
                self._enable_weights_widgets(False, False)
            elif new_index == 1:
                self._enable_weights_widgets(True, False)
            else:
                self._enable_weights_widgets(True, True)
            # Reset the weights table
            if self.current_mode == 2:
                model.clear()
            elif new_index == 2:
                model.data_reset()
            # Reset the scores
            if new_index == 0:
                self.clear_base_scores()
            self.current_mode = new_index

    def _enable_weights_widgets(
        self, enable_base_scores: bool, enable_weights: bool
    ) -> None:
        for widget in self.base_score_widgets:
            widget.setEnabled(enable_base_scores)
        for widget in self.weights_widgets:
            widget.setEnabled(enable_weights)

    def _consolidate_no_scores(self) -> bool:
        self.exam_config.enter_score_mode_none()
        return True

    def _consolidate_base_scores(self) -> bool:
        valid = False
        c_score = self.correct_score.value()
        i_score = self.incorrect_score.value()
        b_score = self.blank_score.value()
        if c_score is not None and c_score > 0:
            if i_score is None:
                i_score = 0
            else:
                i_score = -i_score
            if b_score is None:
                b_score = 0
            else:
                b_score = -b_score
            if i_score >= 0 and b_score >= 0:
                base_scores = scoring.QuestionScores(c_score, i_score, b_score)
                same_weights = self.current_mode == 1
                self.exam_config.set_base_scores(base_scores, same_weights=same_weights)
                valid = True
            else:
                QMessageBox.critical(
                    self,
                    _("Error"),
                    _(
                        "The score for incorrect and blank answers "
                        "cannot be greater than 0."
                    ),
                )
        else:
            QMessageBox.critical(
                self, _("Error"), _("You must enter the score for correct answers.")
            )
        return valid

    def _consolidate_weights(self) -> bool:
        model = cast(ScoreWeightsTableModel, self.weights_table.model())
        valid = model.consolidate()
        if not valid:
            self._show_error_weights()
        return valid



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2021 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#

import unittest

import eyegrade.students as students
import eyegrade.exams as exams


class TestRankStudentIds(unittest.TestCase):
    def test_different_id_len(self):
        # Test for issue #155
        student_list = [
            students.Student("101010101", "Donald Duck", "", "", ""),
            students.Student("202020202", "Marty McFly", "", "", ""),
            students.Student("313131313", "Peter Pan", "", "", ""),
            students.Student("12345678", "Shorter ID", "", "", ""),
            students.Student("1234567890", "Longer ID", "", "", ""),
        ]
        listing = students.GroupListing(students.StudentGroup(1, "G"), [])
        listing.add_students(student_list)
        listings = students.StudentListings()
        listings.add_listing(listing)
        scores = []
        for i in range(10):
            scores.append([])
            for _ in range(10):
                scores[i].append(0.1)
        for i, digit in enumerate(student_list[1].student_id):
            scores[i][int(digit)] = 1.0
        exam = _MockExamForScores(student_list[1].student_id, scores, listings)
        rank = exam.rank_students()


class _MockExamForScores(exams.Exam):
    def __init__(self, detected_id, id_scores, student_listings):
        self.decisions = _MockExamDecisions(detected_id, id_scores)
        self.student_listings = student_listings


class _MockExamDecisions:
    def __init__(self, detected_id, id_scores):
        self.detected_id = detected_id
        self.id_scores = id_scores



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2021 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#

import unittest
import os.path

import eyegrade.exams as exams


class TestExamConfig(unittest.TestCase):
    def _get_test_file_path(self, filename):
        dirname = os.path.dirname(os.path.abspath(__file__))
        return os.path.join(dirname, filename)

    def test_variations(self):
        exam_config = exams.ExamConfig(
            filename=self._get_test_file_path("test-variations.eye")
        )
        self.assertEqual(exam_config.question_variations["A"], [0, 0, 0, 0, 0])
        self.assertEqual(exam_config.question_variations["B"], [2, 1, 1, 2, 0])

    def test_model_variations(self):
        exam_config = exams.ExamConfig(
            filename=self._get_test_file_path("test-model-variations.eye")
        )
        self.assertEqual(exam_config.model_variations["A"], 0)
        self.assertEqual(exam_config.model_variations["B"], 1)



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2021 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <https://www.gnu.org/licenses/>.
#
import gettext
import locale
import platform

from PyQt6.QtWidgets import (
    QCheckBox,
    QDialog,
    QDialogButtonBox,
    QFormLayout,
    QLabel,
    QMessageBox,
    QPushButton,
    QScrollArea,
    QSpinBox,
    QTabWidget,
    QVBoxLayout,
    QComboBox,
)

from PyQt6.QtCore import QTimer, Qt, pyqtSignal, QT_VERSION_STR, PYQT_VERSION_STR

import cv2
import numpy as np

from . import widgets
from .. import utils
from .. import detection

t = gettext.translation("eyegrade", utils.locale_dir(), fallback=True)
_ = t.gettext


class DialogComputeScores(QDialog):
    """Dialog to set the parameters to compute scores automatically.

    Example (replace `parent` by the parent widget):

    dialog = DialogComputeScores(parent)
    max_score, penalize = dialog.exec()

    """

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle(_("Compute default scores"))
        layout = QFormLayout()
        self.setLayout(layout)
        self.score = widgets.InputScore(parent=self)
        self.penalize = QCheckBox(_("Penalize incorrect answers"), self)
        buttons = QDialogButtonBox(
            (
                QDialogButtonBox.StandardButton.Ok
                | QDialogButtonBox.StandardButton.Cancel
            )
        )
        layout.addRow(_("Maximum score"), self.score)
        layout.addRow(_("Penalizations"), self.penalize)
        layout.addRow(buttons)
        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)

    def exec(self):
        """Shows the dialog and waits until it is closed.

        Returns the tuple (max_score, penalize) or (None, None) if the
        user cancels.

        """
        success = False
        score = None
        penalize = None
        while not success:
            result = super().exec()
            if result == QDialog.DialogCode.Accepted:
                if self.score.text():
                    score = self.score.value()
                    if score is not None and score > 0:
                        penalize = self.penalize.checkState() == Qt.CheckState.Checked
                        success = True
                if not success:
                    QMessageBox.critical(self, _("Error"), _("Enter a valid score."))
            else:
                score, penalize = None, None
                success = True
        return (score, penalize)


class DialogCameraSelection(QDialog):
    """Shows a dialog that allows choosing a camera.

    Example (replace `parent` by the parent widget):

    dialog = DialogNewSession(parent)
    values = dialog.exec()

    At the end of the dialog, the chosen camera is automatically
    set in the context object.

    """

    capture_period = 100
    camera_error = pyqtSignal()

    def __init__(self, capture_context, parent):
        """Initializes the dialog.

        `capture_context` is the detection.ExamCaptureContext object
        to be used.

        """
        super().__init__(parent)
        self.capture_context = capture_context
        self.setWindowTitle(_("Select a camera"))
        layout = QVBoxLayout(self)
        self.setLayout(layout)
        self.camview = widgets.CamView((320, 240), self, border=True)
        self.label = QLabel(self)
        self.button = QPushButton(_("Try this camera"))
        self.camera_selector = QSpinBox(self)
        container = widgets.LineContainer(self, self.camera_selector, self.button)
        self.flip_combo = QComboBox(parent=self)
        self.flip_combo.addItem(_("No"))
        self.flip_combo.addItem(_("Horizontally"))
        self.flip_combo.addItem(_("Vertically"))
        self.flip_combo.addItem(_("Both axes"))
        self._init_flip_combo()
        flip_container = widgets.LineContainer(
            self, QLabel(_("Flip image")), self.flip_combo
        )
        self.button.clicked.connect(self._select_camera)
        buttons = QDialogButtonBox(QDialogButtonBox.StandardButton.Ok)
        buttons.accepted.connect(self.accept)
        self.flip_combo.currentIndexChanged.connect(self._flip_changed)
        layout.addWidget(self.camview)
        layout.addWidget(self.label)
        layout.addWidget(container)
        layout.addWidget(flip_container)
        layout.addWidget(buttons)
        self.camera_error.connect(
            self._show_camera_error, type=Qt.ConnectionType.QueuedConnection
        )
        self.timer = None

    def __del__(self):
        if self.timer is not None:
            self.timer.stop()
        self.capture_context.close_camera()

    def exec(self):
        success = self.capture_context.open_camera()
        if success:
            self._update_camera_label()
            self.timer = QTimer(self)
            self.timer.setSingleShot(False)
            self.timer.timeout.connect(self._next_capture)
            self.timer.setInterval(DialogCameraSelection.capture_period)
            self.timer.start()
        else:
            self.camera_error.emit()
        return super().exec()

    def _show_camera_error(self):
        QMessageBox.critical(
            self,
            _("Camera not available"),
            _("Eyegrade has not detected any camera in your system."),
        )
        self.reject()

    def _select_camera(self):
        current_camera = self.capture_context.current_camera_id()
        new_camera = self.camera_selector.value()
        if new_camera != current_camera:
            success = self.capture_context.open_camera(camera_id=new_camera)
            if not success:
                self.camera_error.emit()
            else:
                self._update_camera_label()
                camera_id = self.capture_context.current_camera_id()
                if camera_id != new_camera:
                    QMessageBox.critical(
                        self,
                        _("Camera not available"),
                        _("Camera {0} is not available.").format(new_camera),
                    )
                else:
                    self.flip_combo.setCurrentIndex(0)

    def _update_camera_label(self):
        camera_id = self.capture_context.current_camera_id()
        if camera_id is not None and camera_id >= 0:
            self.label.setText(
                _("<center>Viewing camera: {0}</center>").format(camera_id)
            )
            self.camera_selector.setValue(camera_id)
        else:
            self.label.setText(_("<center>No camera</center>"))

    def _flip_changed(self, index):
        if index == 0:
            transformation = detection.ImageTransformer.IDENTITY
        elif index == 1:
            transformation = detection.ImageTransformer.FLIP_H
        elif index == 2:
            transformation = detection.ImageTransformer.FLIP_V
        elif index == 3:
            transformation = detection.ImageTransformer.FLIP_BOTH
        self.capture_context.image_transformer = detection.ImageTransformer(
            transformation
        )

    def _init_flip_combo(self):
        transformation = self.capture_context.image_transformer.transformation
        if transformation == detection.ImageTransformer.IDENTITY:
            self.flip_combo.setCurrentIndex(0)
        elif transformation == detection.ImageTransformer.FLIP_H:
            self.flip_combo.setCurrentIndex(1)
        elif transformation == detection.ImageTransformer.FLIP_V:
            self.flip_combo.setCurrentIndex(2)
        elif transformation == detection.ImageTransformer.FLIP_BOTH:
            self.flip_combo.setCurrentIndex(3)

    def _next_capture(self):
        if not self.isVisible():
            self.timer.stop()
            self.capture_context.close_camera()
        else:
            image = self.capture_context.capture(resize=(320, 240))
            self.camview.display_capture(image)


class DialogAbout(QDialog):
    """About dialog.

    Example (replace `parent` by the parent widget):

    dialog = DialogAbout(parent)
    values = dialog.exec()

    """

    _tuple_strxfrm = staticmethod(lambda x: locale.strxfrm(x[0]))

    def __init__(self, parent):
        super().__init__(parent)
        self.setWindowTitle(_("About"))
        layout = QVBoxLayout(self)
        self.setLayout(layout)
        buttons = QDialogButtonBox(QDialogButtonBox.StandardButton.Ok)
        buttons.accepted.connect(self.accept)
        tabs = QTabWidget(parent)
        tabs.setDocumentMode(True)
        tabs.addTab(self._create_about_tab(), _("About"))
        tabs.addTab(self._create_developers_tab(), _("Developers"))
        tabs.addTab(self._create_translators_tab(), _("Translators"))
        tabs.addTab(self._create_versions_tab(), _("Software versions"))
        layout.addWidget(tabs)
        layout.addWidget(buttons)

    def _create_about_tab(self):
        text = _(
            """
             <center>
             <p><img src='{0}' width='64'> <br>
             {1} {2} <br>
             (c) 2010-2021 Jesús Arias Fisteus and contributors<br>
             <a href='{3}'>{3}</a> <br>
             <a href='{4}'>{4}</a>

             <p>
             This program is free software: you can redistribute it<br>
             and/or modify it under the terms of the GNU General<br>
             Public License as published by the Free Software<br>
             Foundation, either version 3 of the License, or (at your<br>
             option) any later version.
             </p>
             <p>
             This program is distributed in the hope that it will be<br>
             useful, but WITHOUT ANY WARRANTY; without even the<br>
             implied warranty of MERCHANTABILITY or FITNESS FOR A<br>
             PARTICULAR PURPOSE. See the GNU General Public<br>
             License for more details.
             </p>
             <p>
             You should have received a copy of the GNU General<br>
             Public License along with this program.  If not, see<br>
             <a href='https://www.gnu.org/licenses/gpl.txt'>
             https://www.gnu.org/licenses/gpl.txt</a>.
             </p>
             </center>
             """
        ).format(
            utils.resource_path("logo.svg"),
            utils.program_name,
            utils.version,
            utils.web_location,
            utils.source_location,
        )
        label = QLabel(text)
        label.setOpenExternalLinks(True)
        label.setTextInteractionFlags(
            (
                Qt.TextInteractionFlag.LinksAccessibleByKeyboard
                | Qt.TextInteractionFlag.LinksAccessibleByMouse
                | Qt.TextInteractionFlag.TextBrowserInteraction
                | Qt.TextInteractionFlag.TextSelectableByKeyboard
                | Qt.TextInteractionFlag.TextSelectableByMouse
            )
        )
        return label

    def _create_developers_tab(self):
        text = """<p><b>{0}:</b></p>
                   <ul><li>Jesús Arias Fisteus</li></ul>
                   <p><b>{1}:</b></p>
                   <ul><li>Jonathan Araneda Labarca</li></ul>
                   <p><b>{2}:</b></p>
                   <ul><li>Rodrigo Argüello</li></ul>
                   <p><b>{3}:</b></p>
                   <ul><li>Roberto González</li></ul>
                   """.format(
            _("Lead developers"),
            _("Exam configuration dialogs"),
            _("Manuscript digits recognition"),
            _("Testing and other contributions"),
        )
        label = QLabel(text)
        label.setTextInteractionFlags(
            (
                Qt.TextInteractionFlag.LinksAccessibleByKeyboard
                | Qt.TextInteractionFlag.LinksAccessibleByMouse
                | Qt.TextInteractionFlag.TextBrowserInteraction
                | Qt.TextInteractionFlag.TextSelectableByKeyboard
                | Qt.TextInteractionFlag.TextSelectableByMouse
            )
        )
        scroll_area = QScrollArea(self.parent())
        scroll_area.setWidget(label)
        return scroll_area

    def _create_translators_tab(self):
        translators = [
            (_("Catalan"), ["Jaume Barcelo"]),
            (_("German"), []),
            (_("Galician"), ["Jesús Arias Fisteus"]),
            (_("French"), []),
            (_("Portuguese"), []),
            (_("Spanish"), ["Jesús Arias Fisteus"]),
        ]
        parts = []
        for language, names in sorted(translators, key=DialogAbout._tuple_strxfrm):
            if names:
                parts.append("<p><b>{0}:</b></p>".format(language))
                parts.append("<ul>")
                for name in names:
                    parts.append("<li>{0}</li>".format(name))
                parts.append("</ul>")
        label = QLabel("".join(parts))
        label.setTextInteractionFlags(
            (
                Qt.TextInteractionFlag.LinksAccessibleByKeyboard
                | Qt.TextInteractionFlag.LinksAccessibleByMouse
                | Qt.TextInteractionFlag.TextBrowserInteraction
                | Qt.TextInteractionFlag.TextSelectableByKeyboard
                | Qt.TextInteractionFlag.TextSelectableByMouse
            )
        )
        scroll_area = QScrollArea(self.parent())
        scroll_area.setWidget(label)
        return scroll_area

    def _create_versions_tab(self):
        align = "style='text-align: right'"
        text = f"""
            <center>
            <table>
            <tr><td>{utils.program_name}</td><td {align}>{utils.version}</td></tr>
            <tr><td>Python</td><td {align}>{platform.python_version()}</td></tr>
            <tr><td>Qt</td><td {align}>{QT_VERSION_STR}</td></tr>
            <tr><td>PyQt</td><td {align}>{PYQT_VERSION_STR}</td></tr>
            <tr><td>OpenCV</td><td {align}>{cv2.version.opencv_version}</td></tr>
            <tr><td>NumPy</td><td {align}>{np.__version__}</td></tr>
            </table>
            </center>
             """
        label = QLabel(text)
        label.setTextInteractionFlags(
            (
                Qt.TextInteractionFlag.LinksAccessibleByKeyboard
                | Qt.TextInteractionFlag.LinksAccessibleByMouse
                | Qt.TextInteractionFlag.TextBrowserInteraction
                | Qt.TextInteractionFlag.TextSelectableByKeyboard
                | Qt.TextInteractionFlag.TextSelectableByMouse
            )
        )
        return label



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2021 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <https://www.gnu.org/licenses/>.
#

import gettext

from PyQt6.QtGui import QColor

from .. import utils

t = gettext.translation("eyegrade", utils.locale_dir(), fallback=True)
_ = t.gettext


class FileNameFilters:
    exam_config = _("Exam configuration (*.eye)")
    session_db = _("Eyegrade session (*.eyedb)")
    student_list = _("Student list (*.xlsx *.csv *.tsv *.txt *.lst *.list)")
    xlsx_file = _("Excel spreadsheet (*.xlsx)")
    csv_file = _("Data file (*.csv *.tsv)")


class Colors:
    eyegrade_blue = QColor(32, 73, 124)



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2021 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <https://www.gnu.org/licenses/>.
#

import abc
import re
import csv
import itertools
import enum
import io
from typing import Iterator, Optional, Iterable, Union

import openpyxl

from . import utils

re_email = r"^[a-zA-Z0-9._%-\+]+@[a-zA-Z0-9._%-]+.[a-zA-Z]{2,6}$"
_re_email = re.compile(re_email)
_re_student_id = re.compile(r"^[0-9]+$")


class Student:
    student_id: str
    full_name: Optional[str]
    first_name: Optional[str]
    last_name: Optional[str]
    email: Optional[str]
    db_id: Optional[int]
    group_id: Optional[int]
    sequence_num: Optional[int]
    is_in_database: bool

    def __init__(
        self,
        student_id: str,
        full_name: Optional[str],
        first_name: Optional[str],
        last_name: Optional[str],
        email: Optional[str],
        db_id: Optional[int] = None,
        group_id: Optional[int] = None,
        sequence_num: Optional[int] = None,
        is_in_database: bool = False,
    ) -> None:
        if full_name and (first_name or last_name):
            raise ValueError("Full name incompatible with first / last name")
        self.db_id = db_id
        self.student_id = student_id
        self.full_name = full_name
        self.first_name = first_name
        self.last_name = last_name
        self.email = email
        self.group_id = group_id
        self.sequence_num = sequence_num
        self.is_in_database = is_in_database
        self.is_duplicate = False

    @property
    def name(self) -> str:
        if self.full_name:
            return self.full_name
        elif self.last_name:
            if self.first_name:
                return "{0} {1}".format(self.first_name, self.last_name)
            else:
                return self.last_name
        elif self.first_name:
            return self.first_name
        else:
            return ""

    @property
    def last_comma_first_name(self) -> str:
        if self.last_name:
            if self.first_name:
                return "{0}, {1}".format(self.last_name, self.first_name)
            else:
                return self.last_name
        else:
            return self.name

    @property
    def id_and_name(self) -> str:
        if self.name:
            return " ".join((self.student_id, self.name))
        else:
            return self.student_id

    @property
    def name_or_id(self) -> str:
        if self.name:
            return self.name
        elif self.student_id:
            return self.student_id
        else:
            return ""

    def __lt__(self, other: "Student") -> bool:
        return self.student_id < other.student_id

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Student):
            return self.student_id == other.student_id
        else:
            return False

    def __str__(self) -> str:
        return "student: " + self.id_and_name


class StudentGroup:
    identifier: int
    name: str

    def __init__(self, identifier: int, name: str) -> None:
        self.identifier = identifier
        self.name = name

    def __str__(self) -> str:
        return "Group #{0.identifier} ({0.name})".format(self)


class GroupListing:
    group: StudentGroup
    students: list[Student]
    parent: Optional["StudentListings"]
    _students_dict: dict[str, Student]

    def __init__(self, group: StudentGroup, students: Iterable[Student]) -> None:
        self.group = group
        self.students = list(students)
        self.parent = None
        self._students_dict = {s.student_id: s for s in students}

    def student(self, student_id: str) -> Optional[Student]:
        return self._students_dict.get(student_id, None)

    def add_students(self, student_list: list[Student]) -> None:
        if student_list:
            if self.parent is not None:
                duplicates = self.parent.find_duplicates(student_list)
            else:
                duplicates = self.find_duplicates(student_list)
            if not duplicates:
                if student_list[0].sequence_num is None:
                    self._update_sequence_num(student_list)
                self.students.extend(student_list)
                for student in student_list:
                    student.group_id = self.group.identifier
                self._students_dict.update({s.student_id: s for s in student_list})
            else:
                raise DuplicateStudentIdException(duplicates)

    def remove_students(self, students: Iterable[Student]) -> None:
        for student in students:
            if student.student_id in self._students_dict:
                del self._students_dict[student.student_id]
                self.students.remove(student)

    def rename(self, new_name: str) -> None:
        self.group.name = new_name

    def find_duplicates(self, students: Iterable[Student]) -> list[Student]:
        non_duplicates, duplicates = _duplicate_student_ids(students)
        duplicates.extend([s for s in non_duplicates if s.student_id in self])
        return duplicates

    def __len__(self) -> int:
        return len(self.students)

    def __getitem__(self, key: int) -> Student:
        return self.students[key]

    def __iter__(self) -> Iterator[Student]:
        return iter(self.students)

    def __contains__(self, student_id: str) -> bool:
        return student_id in self._students_dict

    def __str__(self) -> str:
        return "GroupListing({}, {} students)".format(self.group, len(self.students))

    def _update_sequence_num(self, students: list[Student]) -> None:
        if self.students:
            highest_sequence_num = max(
                s.sequence_num for s in self.students if s.sequence_num is not None
            )
            first_num = highest_sequence_num + 1
        else:
            first_num = 1
        for i, student in enumerate(students):
            student.sequence_num = first_num + i


class StudentListings:
    listings: list[GroupListing]
    max_group_id: int

    def __init__(self):
        self.listings = []
        self.max_group_id = -1

    def add_listing(self, listing: GroupListing) -> None:
        duplicates = self.find_duplicates(listing.students)
        if not duplicates:
            listing.parent = self
            self.listings.append(listing)
            group_id = listing.group.identifier
            if group_id > self.max_group_id:
                self.max_group_id = group_id
        else:
            raise DuplicateStudentIdException(duplicates)

    def create_listing(self, group: StudentGroup) -> GroupListing:
        if group.identifier is None:
            group.identifier = self.max_group_id + 1
        listing = GroupListing(group, [])
        self.add_listing(listing)
        return listing

    def remove_at(self, index: int) -> None:
        del self.listings[index]

    def iter_students(self) -> Iterator[Student]:
        return itertools.chain(*self.listings)

    def sorted_students(self, key=lambda x: x.student_id) -> list[Student]:
        return sorted([student for student in self.iter_students()], key=key)

    def student(self, student_id: str) -> Optional[Student]:
        student = None
        for listing in self.listings:
            student = listing.student(student_id)
            if student is not None:
                break
        return student

    def listing_by_group_id(self, group_id: int) -> GroupListing:
        for listing in self.listings:
            if listing.group.identifier == group_id:
                return listing
        raise KeyError(group_id)

    def find_duplicates(self, students: Iterable[Student]) -> list[Student]:
        non_duplicates, duplicates = _duplicate_student_ids(students)
        duplicates.extend([s for s in non_duplicates if s.student_id in self])
        return duplicates

    def __len__(self) -> int:
        return len(self.listings)

    def __getitem__(self, key: int) -> GroupListing:
        return self.listings[key]

    def __contains__(self, student_id: str) -> bool:
        for listing in self.listings:
            if student_id in listing:
                return True
        return False

    def __str__(self) -> str:
        return "StudentListings({} groups: {})".format(
            len(self.listings), [len(listing) for listing in self.listings]
        )


def _duplicate_student_ids(
    students: Iterable[Student],
) -> tuple[list[Student], list[Student]]:
    non_duplicates = []
    duplicates = []
    student_ids_set = set()
    for student in students:
        if student.student_id not in student_ids_set:
            student_ids_set.add(student.student_id)
            non_duplicates.append(student)
        else:
            duplicates.append(student)
    return non_duplicates, duplicates


class CantRemoveGroupException(utils.EyegradeException):
    def __init__(self, message: str):
        super().__init__(message)


class DuplicateStudentIdException(utils.EyegradeException):
    def __init__(self, duplicates: Iterable[Student]):
        super().__init__("Some ids are already in the student listings")
        self.duplicates = duplicates


class StudentReader:
    file_name: str
    column_map: Optional["StudentColumnMap"]
    iterator: Iterator[Union[tuple[str], list[str]]]

    def __init__(
        self, file_name: str, column_map: Optional["StudentColumnMap"] = None
    ) -> None:
        self.file_name = file_name
        self.column_map = column_map
        # To be overwritten by subclasses:
        self.iterator = iter([])

    @staticmethod
    def create(file_name: str) -> "StudentReader":
        if file_name.endswith(".xlsx"):
            return XLSXStudentReader(file_name)
        else:
            return CSVStudentReader(file_name)

    @abc.abstractmethod
    def __enter__(self) -> "StudentReader": ...

    @abc.abstractmethod
    def __exit__(self, exception_type, exception_value, traceback) -> None: ...

    def students(self) -> Iterator[Student]:
        first_line = True
        for row in self.iterator:
            if not StudentReader._row_is_empty(row):
                if self.column_map is None:
                    self.column_map = StudentColumnMap.guess_map(row)
                    if not self.column_map.is_valid():
                        if first_line:
                            first_line = False
                            self.column_map = None
                            continue
                        else:
                            raise utils.EyegradeException("", key="error_student_list")
                try:
                    student = self.column_map.student(row)
                    yield student
                except utils.EyegradeException:
                    if not first_line:
                        raise
                first_line = False

    @staticmethod
    def _row_is_empty(row: Iterable[Optional[str]]) -> bool:
        for element in row:
            if element is not None and element != "":
                return False
        return True


class CSVStudentReader(StudentReader):
    file: Optional[io.TextIOWrapper]

    def __init__(self, file_name: str) -> None:
        super().__init__(file_name)
        self.file = None

    def __enter__(self) -> StudentReader:
        dialect: type[csv.Dialect]
        self.file = open(self.file_name, newline="")
        file_sample = self.file.read(1024)
        if "\t" in file_sample:
            # The sniffer doesn't guess correctly when a TSV file
            # contains names with commas
            dialect = csv.excel_tab
        else:
            try:
                dialect = csv.Sniffer().sniff(file_sample)
            except csv.Error:
                dialect = csv.excel_tab
        self.file.seek(0)
        self.iterator = csv.reader(self.file, dialect=dialect)
        return self

    def __exit__(self, exception_type, exception_value, traceback) -> None:
        if self.file is not None:
            self.file.close()


class XLSXStudentReader(StudentReader):
    workbook: openpyxl.Workbook

    def __init__(self, file_name: str) -> None:
        super().__init__(file_name)
        self.workbook = None

    def __enter__(self) -> StudentReader:
        self.workbook = openpyxl.load_workbook(self.file_name, read_only=True)
        self.iterator = self.iter_rows(self.workbook.active)
        return self

    def __exit__(self, exception_type, exception_value, traceback) -> None:
        self.workbook.close()

    def iter_rows(
        self, work_sheet: openpyxl.worksheet.worksheet.Worksheet
    ) -> Iterator[tuple]:
        for row in work_sheet.iter_rows():
            yield tuple(cell.value for cell in row)


def read_students(file_name: str) -> list[Student]:
    """Reads the list of students from a file.

    Formats allowed: CSV-formatted file (tab-separated) and Excel 2010 (.xslx)

    Returns the results as a list of Student objects.

    """
    with StudentReader.create(file_name) as reader:
        return list(reader.students())


class StudentColumn(enum.Enum):
    ID = 1
    FULL_NAME = 2
    FIRST_NAME = 3
    LAST_NAME = 4
    NAME = 5
    EMAIL = 6
    SEQUENCE_NUM = 7
    UNKNOWN = 8


ATTR_NAME = {
    StudentColumn.ID: "student_id",
    StudentColumn.FULL_NAME: "full_name",
    StudentColumn.FIRST_NAME: "first_name",
    StudentColumn.LAST_NAME: "last_name",
    StudentColumn.NAME: "name",
    StudentColumn.EMAIL: "email",
    StudentColumn.SEQUENCE_NUM: "sequence_num",
    StudentColumn.UNKNOWN: "-",
}


class StudentColumnMap:
    student_column: list[StudentColumn]

    def __init__(self, num_columns=None, columns=None):
        if not ((num_columns is None) ^ (columns is None)):
            raise ValueError("num_columns or columns required, but not both")
        if num_columns is not None:
            self.columns = [StudentColumn.UNKNOWN] * num_columns
        else:
            self.columns = list(columns)

    def set_column(self, index: int, column: StudentColumn) -> bool:
        if column not in self.columns:
            self.columns[index] = column
            return True
        else:
            return False

    def resolve(self) -> None:
        # Identify first name / last name / full name columns
        # They are marked as unknown until now
        num_columns = len(self.columns)
        for i in range(num_columns):
            if self.columns[i] == StudentColumn.UNKNOWN:
                if i == num_columns - 1 or self.columns[i + 1] != StudentColumn.UNKNOWN:
                    self.columns[i] = StudentColumn.FULL_NAME
                else:
                    self.columns[i] = StudentColumn.FIRST_NAME
                    self.columns[i + 1] = StudentColumn.LAST_NAME
                break
        for i, col in reversed(list(enumerate(self.columns))):
            if col != StudentColumn.UNKNOWN:
                break
        self.columns = self.columns[: i + 1]

    def is_valid(self) -> bool:
        return StudentColumn.ID in self.columns

    def student(self, row: Union[tuple[str], list[str]]) -> Student:
        num_columns = len(self.columns)
        if len(row) < num_columns:
            raise utils.EyegradeException(
                "Row with not enough columns", key="error_student_list"
            )
        student = Student("", "", "", "", "")
        for i, item in enumerate(row[:num_columns]):
            if self.columns[i] != StudentColumn.UNKNOWN:
                value = str(item)
                self._check_value(self.columns[i], value)
                attr_name = ATTR_NAME[self.columns[i]]
                setattr(student, attr_name, value)
        return student

    def data(self, index: int, student: Student) -> str:
        column = self.columns[index]
        if column != StudentColumn.UNKNOWN:
            attr_name = ATTR_NAME[column]
            return getattr(student, attr_name)
        else:
            return ""

    def normalize(self) -> "StudentColumnMap":
        normal_order = [
            StudentColumn.ID,
            StudentColumn.FIRST_NAME,
            StudentColumn.LAST_NAME,
            StudentColumn.FULL_NAME,
            StudentColumn.EMAIL,
        ]
        reordered_columns = [
            column for column in normal_order if column in self.columns
        ]
        return StudentColumnMap(columns=reordered_columns)

    def to_full_name(self) -> "StudentColumnMap":
        # It will raise ValueError if first or last name aren't present
        index_first = self.columns.index(StudentColumn.FIRST_NAME)
        index_last = self.columns.index(StudentColumn.LAST_NAME)
        new_columns = list(self.columns)
        if index_first < index_last:
            new_columns[index_first] = StudentColumn.FULL_NAME
            del new_columns[index_last]
        else:
            new_columns[index_last] = StudentColumn.FULL_NAME
            del new_columns[index_first]
        return StudentColumnMap(columns=new_columns)

    def __str__(self) -> str:
        return (
            "StudentColumnMap <"
            + ", ".join(ATTR_NAME[column] for column in self.columns)
            + ">"
        )

    def __len__(self) -> int:
        return len(self.columns)

    def __getitem__(self, index: int) -> StudentColumn:
        return self.columns[index]

    def __contains__(self, column: StudentColumn) -> bool:
        return column in self.columns

    @staticmethod
    def guess_map(row: Union[tuple[str], list[str]]) -> "StudentColumnMap":
        column_map = StudentColumnMap(num_columns=len(row))
        for i, item in enumerate(row):
            value = str(item)
            if _re_student_id.match(value):
                column = StudentColumn.ID
            elif _re_email.match(value):
                column = StudentColumn.EMAIL
            else:
                column = StudentColumn.UNKNOWN
            column_map.set_column(i, column)
        column_map.resolve()
        return column_map

    def _check_value(self, column: StudentColumn, value: str) -> None:
        if column == StudentColumn.ID:
            if not _re_student_id.match(value):
                raise utils.EyegradeException(
                    "Wrong id in student list: " + value, key="error_student_list"
                )
        elif column == StudentColumn.EMAIL:
            if not _re_email.match(value):
                raise utils.EyegradeException(
                    "Wrong email in student list: " + value, key="error_student_list"
                )


utils.EyegradeException.register_error(
    "error_student_list", "The syntax of the student list isn't correct."
)



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2021 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <https://www.gnu.org/licenses/>.
#

import re
import copy
import sys
import subprocess
import os

from .. import utils
from .. import exams


PARAM_MIN_NUM_QUESTIONS = 1

# For formatting questions
PARAM_TABLE_SEP = 0.05
PARAM_TABLE_MARGIN = 0.1

# Numbers of questions in which the number of tables is changed
PARAM_TABLE_LMITS = [8, 24, 55]
RE_SPLIT_TEMPLATE = re.compile("{{([^{}]+)}}")

# Register user-friendly error messages
utils.EyegradeException.register_error(
    "incoherent_exam_config",
    "The exam you are attempting to create is not compatible\n"
    "with the already existing .eye exam configuration file.\n"
    "This happens, for example, when the configuration file\n"
    "contains more or less questions than the exam you are now\n"
    "creating. If you really want to discard the previous .eye file,\n"
    "use the --force option, or just remove the file manually.",
)
utils.EyegradeException.register_error(
    "incoherent_num_tables",
    "The specified number of tables and the exam dimensions are not\n"
    "compatible. The exam dimensions implicitly specify the number of tables.\n"
    "Therefore, there is no need to explicitly specify the number of tables.",
    "Incoherent number of tables.",
)
utils.EyegradeException.register_error(
    "bad_model_value",
    "A model must be represented by an uppercase English letter (A-Z).\n"
    "You can also create an unshuffled version of the exam with the\n"
    "special '0' model.",
    "Bad model value.",
)
utils.EyegradeException.register_error(
    "too_few_questions",
    short_message="At least %d question(s) needed" % PARAM_MIN_NUM_QUESTIONS,
)
utils.EyegradeException.register_error(
    "too_few_choices", short_message="At least 2 choices per question needed"
)
utils.EyegradeException.register_error(
    "too_many_tables",
    "There cannot be less than two questions per table.",
    "There are too many tables for such a few questions",
)
utils.EyegradeException.register_error(
    "latex_not_found",
    "Install LaTeX and make sure it is in your system's PATH variable.",
    "The command pdflatex was not found.",
)
utils.EyegradeException.register_error(
    "latex_compile_error",
    (
        "LaTeX failed to compile the exam.\n"
        "Check the error message above or the log for details.\n"
        "LaTeX source files (.tex extension) will not be removed\n"
        "so that you can inspect them."
    ),
    "LaTeX failed to compile the exam.",
)
utils.EyegradeException.register_error(
    "no_exam_config_variations",
    "Setting an output file name is mandatory "
    "when a variation number is explicitly set. "
    "Use the -o option if calling this from the command line.",
)


class ExamMaker:
    def __init__(
        self,
        num_questions,
        num_choices,
        template_filename,
        output_file,
        variables,
        exam_config_filename,
        num_tables=0,
        dimensions=None,
        table_width=None,
        table_height=None,
        table_scale=1.0,
        id_box_width=None,
        force_config_overwrite=False,
        scores=None,
        points_words=None,
        left_to_right_numbering=False,
        survey_mode=False,
    ):
        """
        Class able to create exams. One object is enough for all models.

        """
        self.num_questions = num_questions
        self.num_choices = num_choices
        template = utils.read_file(template_filename)
        self.parts = RE_SPLIT_TEMPLATE.split(template)
        self.left_to_right_numbering = left_to_right_numbering
        self.survey_mode = survey_mode
        self.output_file = output_file
        self.exam_questions = None
        template_id_label, template_id_num_digits = id_num_digits(self.parts)
        if "student_id_label" in variables:
            id_label = variables["student_id_label"]
        else:
            id_label = template_id_label
        if "student_id_length" in variables:
            self.id_num_digits = variables["student_id_length"]
        else:
            self.id_num_digits = template_id_num_digits
        # The ID box is part of the answer table replacement
        # when it does not appear as a separate key in the template.
        # The latter is legacy behaviour kept for backwards compatibility.
        if template_id_label is None and self.id_num_digits > 0:
            self.id_box_with_answer_table = True
            if id_label is None:
                id_label = "ID"
        else:
            self.id_box_with_answer_table = False
        self.exam_config_filename = exam_config_filename
        if num_tables > 0 and dimensions is not None and len(dimensions) != num_tables:
            raise utils.EyegradeException("", key="incoherent_num_tables")
        if dimensions is not None:
            self.dimensions = dimensions
        else:
            self.dimensions = compute_table_dimensions(
                num_questions, num_choices, num_tables
            )
        self.table_scale = table_scale
        self.id_box_width = id_box_width
        if table_width is None and table_height is None:
            (
                self.table_width,
                self.table_height,
                self.id_box_width,
            ) = self._compute_table_size()
        else:
            self.table_width = table_width
            self.table_height = table_height
        if self.exam_config_filename is not None:
            if not force_config_overwrite:
                self._load_exam_config()
            else:
                self._new_exam_config()
        else:
            self.exam_config = None
        if scores is not None:
            variables["score_correct"] = scores.format_correct_score()
            variables["score_incorrect"] = scores.format_incorrect_score()
            if self.exam_config is not None:
                self.exam_config.set_base_scores(scores)
        if points_words is not None:
            variables[
                "score_correct"
            ] += f" {points_words.for_score(scores.correct_score)}"
            variables[
                "score_incorrect"
            ] += f" {points_words.for_score(scores.incorrect_score)}"
        self._load_replacements(variables, id_label)
        self.empty_variables = []

    def set_exam_questions(self, exam):
        if exam.num_questions() != self.num_questions:
            raise Exception("Incorrect number of questions")
        self.exam_questions = exam

    def create_exam(
        self,
        model,
        shuffle,
        variation=None,
        with_solution=False,
        produce_pdf=False,
        keep_question_order=False,
        keep_tex=False,
    ):
        """Creates a new exam.

        'shuffle' must be a boolean. If True, the exam is shuffled
        again even if it was previously shuffled. If False, it is
        only shuffled if it was not previously shuffled.

        """
        if not self.survey_mode:
            if (
                model is None
                or len(model) != 1
                or ((ord(model) < 65 or ord(model) > 90) and model != "0")
            ):
                raise utils.EyegradeException("", "bad_model_value")
        else:
            if model is None:
                model = "A"
        replacements = copy.copy(self.replacements)
        answer_table = create_answer_table(
            self.dimensions,
            model,
            self.survey_mode,
            self.table_width,
            self.table_height,
            self.left_to_right_numbering,
        )
        if self.id_box_with_answer_table:
            answer_table = replacements["id-box"] + answer_table
        if self.exam_config is not None:
            if self.exam_config.dimensions == []:
                self.exam_config.dimensions = self.dimensions
            if model != "0" and model not in self.exam_config.models:
                self.exam_config.models.append(model)
        if variation is not None:
            if self.exam_config is not None:
                self.exam_config.model_variations[model] = variation
            else:
                raise utils.EyegradeException("", key="no_exam_config_variations")
        elif self.exam_config is not None:
            variation = self.exam_config.get_model_variation(model)
        if self.exam_questions is not None:
            if model != "0" and not self.survey_mode:
                if (
                    self.exam_config is None
                    or model not in self.exam_config.permutations
                    or (model in self.exam_config.permutations and shuffle)
                ):
                    self.exam_questions.shuffle(
                        model,
                        variation=variation,
                        keep_question_order=keep_question_order,
                    )
                    if self.exam_config is not None:
                        (
                            solutions,
                            permutations,
                        ) = self.exam_questions.solutions_and_permutations(model)
                        self.exam_config.solutions[model] = solutions
                        self.exam_config.permutations[model] = permutations
                        self.exam_config.question_variations[model] = (
                            self.exam_questions.selected_variations(model)
                        )
                else:
                    permutations = self.exam_config.permutations[model]
                    self.exam_questions.set_permutation(model, permutations)
                    if model in self.exam_config.question_variations:
                        variations = self.exam_config.question_variations[model]
                        self.exam_questions.select_variations(model, variations)
            replacements["questions"] = format_questions(
                self.exam_questions, model, with_solution
            )
        replacements["answer-table"] = answer_table
        replacements["model"] = model
        replacements["declarations"] = latex_declarations(with_solution)
        replacements["variation"] = "0"
        if model != "0":
            if variation is not None:
                replacements["variation"] = str(variation + 1)

        # Replacement keys are in odd positions of self.parts
        replaced = len(self.parts) * [None]
        replaced[::2] = self.parts[::2]
        replaced[1::2] = [self._replace(key, replacements) for key in self.parts[1::2]]
        exam_text = "".join(replaced)
        if self.output_file == sys.stdout:
            utils.write_to_stdout(exam_text)
            produced_filename = None
        else:
            if not self.survey_mode:
                produced_filename = self.output_file % model
            else:
                produced_filename = self.output_file
            utils.write_file(produced_filename, exam_text)
            if produce_pdf:
                success, output, produced_filename = compile_latex(
                    produced_filename, keep_tex=keep_tex
                )
                if not success:
                    raise utils.EyegradeException(output, key="latex_compile_error")
        return produced_filename

    def save_exam_config(self):
        if self.exam_config is not None:
            self.exam_config.save(self.exam_config_filename)

    def _load_exam_config(self):
        if self.exam_config_filename is not None:
            try:
                self.exam_config = exams.ExamConfig(self.exam_config_filename)
                if self.num_questions != self.exam_config.num_questions:
                    raise utils.EyegradeException(
                        "Incoherent number of questions", key="incoherent_exam_config"
                    )
                if self.id_num_digits != self.exam_config.id_num_digits:
                    raise utils.EyegradeException(
                        "Incoherent configuration of id box",
                        key="incoherent_exam_config",
                    )
                if self.dimensions != self.exam_config.dimensions:
                    raise utils.EyegradeException(
                        "Incoherent table dimensions", key="incoherent_exam_config"
                    )
                if (
                    self.left_to_right_numbering
                    != self.exam_config.left_to_right_numbering
                ):
                    raise utils.EyegradeException(
                        "Incoherent question numbering", key="incoherent_exam_config"
                    )
                if self.survey_mode != self.exam_config.survey_mode:
                    raise utils.EyegradeException(
                        "Incoherent survey mode value", key="incoherent_exam_config"
                    )
            except IOError:
                self._new_exam_config()

    def _new_exam_config(self):
        self.exam_config = exams.ExamConfig()
        self.exam_config.num_questions = self.num_questions
        self.exam_config.id_num_digits = self.id_num_digits
        self.exam_config.left_to_right_numbering = self.left_to_right_numbering
        self.exam_config.survey_mode = self.survey_mode

    def _compute_table_size(self):
        """Computes a size for the table such as it is more or less square.

        Some facts the function is based on:

        - In Latex, the proportion between the witdh and the height of a
        cell will be 1.55.

        - Each table has an extra column for the question number.

        - There are one extra row at the top, and two at the bottom.

        - The objective is making the lengths of the horizontal lines be
        close to the length of the vertical lines.

        - When resizing the table, with double width proportions are 1:1.

        """
        num_cols = self.dimensions[0][0] * len(self.dimensions)
        num_rows = max([d[1] for d in self.dimensions])
        extra_cols = len(self.dimensions)
        extra_rows = 3
        actual_ratio = 1.55 * num_cols / num_rows
        desired_ratio = (1 + 1.0 * extra_cols / num_cols) / (
            1 + 1.0 * extra_rows / num_rows
        )
        if actual_ratio > 1.35 * desired_ratio:
            ratio = 1.35 * desired_ratio
        elif actual_ratio < 0.74 * desired_ratio:
            ratio = 0.74 * desired_ratio
        else:
            ratio = actual_ratio
        height = 0.3 * num_rows
        if height < 3:
            height = 3.0
        elif height > 4.6:
            height = 4.6
        height = height * self.table_scale
        width = 2 * height * ratio
        if self.id_num_digits > 0 and self.id_box_width is None:
            actual_id_width = 0.75 * self.id_num_digits
            if actual_id_width < 0.66 * width:
                id_width = 0.66 * width
            elif actual_id_width > 1.5 * width:
                id_width = 1.5 * width
            else:
                id_width = None
        else:
            id_width = None
        return width, height, id_width

    def _load_replacements(self, variables, id_label):
        self.replacements = copy.copy(variables)
        self.replacements["id-box"] = create_id_box(
            id_label, self.id_num_digits, self.id_box_width
        )
        self.replacements["questions"] = ""

    def _replace(self, key, replacements):
        if key in replacements:
            if not replacements[key] and key not in self.empty_variables:
                self.empty_variables.append(key)
                return ""
            else:
                return replacements[key]
        elif key.startswith("id-box"):
            return replacements["id-box"]
        else:
            raise Exception("Unknown replacement key: " + key)


def check_latex():
    with open(os.devnull, "w") as devnull:
        try:
            subprocess.check_call(
                ["pdflatex", "-version"], stdout=devnull, stderr=devnull
            )
        except (subprocess.CalledProcessError, FileNotFoundError):
            success = False
        else:
            success = True
    return success


def compile_latex(latex_file, keep_tex=False):
    directory, name = os.path.split(latex_file)
    base_name = os.path.splitext(name)[0]
    with utils.change_dir(directory):
        try:
            output = subprocess.check_output(
                ["pdflatex", "-interaction=nonstopmode", name],
                stderr=subprocess.STDOUT,
                text=True,
            )
        except subprocess.CalledProcessError as exc:
            output = exc.output
            success = False
            produced_filename = None
        except OSError:
            success = False
            raise utils.EyegradeException("", key="latex_not_found")
        else:
            success = True
            produced_filename = os.path.join(directory, base_name + ".pdf")
        finally:
            to_remove = [base_name + ".aux"]
            if success:
                to_remove.append(base_name + ".log")
                if not keep_tex:
                    to_remove.append(name)
            for filename in to_remove:
                if os.path.isfile(filename):
                    os.remove(filename)
    return success, output, produced_filename


def latex_declarations(with_solution):
    """Returns the list of declarations to be set in the preamble
    of the LaTeX output.

    """
    data = [
        r"\usepackage{graphicx}",
        r"\usepackage{fancyvrb}",
        r"\usepackage{enumerate}",
        r"\usepackage{color}",
        r"\definecolor{lightgray}{rgb}{1, 1, 1}",
        r"\newcommand{\light}[1]{\textcolor{lightgray}{#1}}",
        r"\definecolor{hidden}{rgb}{1, 1, 1}",
        r"\newcommand{\hidden}[1]{\textcolor{hidden}{#1}}",
        r"\newif\ifsolutions",
    ]
    if with_solution:
        data.append(r"\solutionstrue")
    else:
        data.append(r"\solutionsfalse")
    return "\n".join(data)


def create_answer_table(
    dimensions,
    model,
    survey_mode,
    table_width=None,
    table_height=None,
    left_to_right_numbering=False,
):
    """Returns a string with the answer tables of the answer sheet.

    Tables are LaTeX-formatted. 'dimensions' specifies the geometry
    of the tables. 'model' is a one letter string with the name of
    the model, or '0' for the un-shuffled exam. 'table_width' is
    the desired width of the answer table, in cm. None for the
    default width. 'table_height' is the desired height in cm. None
    for the default. If only one of height and width is defined,
    the other will keep the aspect ratio. 'left_to_right_numbering'
    set to True makes cell numbers to grow from left to right
    instead of from up to bottom. The default is False (up to
    bottom).

    """
    if not dimensions:
        raise Exception("No tables defined in dimensions")
    compact = True
    num_choices = dimensions[0][0]
    num_tables = len(dimensions)
    for dimension in dimensions:
        if dimension[0] != num_choices:
            raise utils.EyegradeException("", "same_num_choices")
    if model != "0":
        bits = utils.encode_model(model, num_tables, num_choices)
    else:
        bits = [False] * num_tables * num_choices
    bits_rows = _create_infobits(bits, num_tables, num_choices, survey_mode)
    tables, question_numbers = table_geometry(dimensions)
    rows = _table_top(num_tables, num_choices, compact, table_width, table_height)
    for i, row_geometry in enumerate(tables):
        rows.append(_horizontal_line(row_geometry, num_choices, compact))
        rows.append(
            _build_row(
                i,
                row_geometry,
                question_numbers,
                num_choices,
                bits_rows,
                compact,
                left_to_right_numbering,
            )
        )
    rows.append(r"\end{tabular}")
    if table_width is not None or table_height is not None:
        rows.append("}")
    rows.append(r"\end{center}")
    return "\n".join(rows)


def create_id_box(label, num_digits, box_width=None):
    """Creates the ID box given a label to show and number of digits."""
    parts = [r"\begin{center}", r"\Large"]
    if box_width is not None:
        parts.append(r"\resizebox{%fcm}{!}{" % box_width)
    parts.append(r"\begin{tabular}{l|" + num_digits * "p{3mm}|" + "}")
    parts.append(r"\cline{2-%d}" % (1 + num_digits))
    parts.append(r"\textbf{%s}: " % label + num_digits * "& " + r"\\")
    parts.append(r"\cline{2-%d}" % (1 + num_digits))
    parts.append(r"\end{tabular}")
    if box_width is not None:
        parts.append("}")
    parts.append(r"\end{center}")
    return "\n".join(parts)


def id_num_digits(parts):
    """Returns the tuple (label, number of digits) for the ID box.

    Receives the splitted text of the template for the exam, which
    may contain a key like 'id-box(9,NIA)' in an odd position, in
    which NIA is the label and 9 the number of digits. If the key
    does not exist in the template, returns (0, None).

    """
    # Replacement keys are in odd positions of the list
    for part in parts[1::2]:
        if part.startswith("id-box"):
            data = part[7:-1].split(",")
            # data[0] is num_digits; data[1] is label
            return data[1], int(data[0])
    return None, 0


def compute_table_dimensions(num_questions, num_choices, num_tables):
    """Computes and returns dimensions for answer tables.

    The result is a list of tuples in which each element contains
    the pair (num_cols, num_rows) for each table.

    """
    if num_questions < PARAM_MIN_NUM_QUESTIONS:
        raise utils.EyegradeException("", key="too_few_questions")
    if num_choices < 2:
        raise utils.EyegradeException("", key="too_few_choices")
    if num_tables <= 0:
        num_tables = _choose_num_tables(num_questions)
    elif num_tables * 2 > num_questions:
        raise utils.EyegradeException("", key="too_many_tables")
    dimensions = []
    rows_per_table, extra_rows = divmod(num_questions, num_tables)
    for i in range(0, num_tables):
        if i < extra_rows:
            num_rows = rows_per_table + 1
        else:
            num_rows = rows_per_table
        dimensions.append((num_choices, num_rows))
    return dimensions


def table_geometry(dimensions):
    num_cols = [table[0] for table in dimensions]
    num_rows = 2 + max([table[1] for table in dimensions])
    tables = []
    for i in range(0, num_rows):
        row = []
        for j, num_choices in enumerate(num_cols):
            if i < dimensions[j][1]:
                row.append(num_choices)
            elif i == dimensions[j][1]:
                row.append(-1)
            elif i == dimensions[j][1] + 1:
                row.append(-2)
            else:
                row.append(0)
        tables.append(row)
    question_numbers = [1]
    for i in range(0, len(dimensions) - 1):
        question_numbers.append(question_numbers[-1] + dimensions[i][1])
    return tables, question_numbers


def _choose_num_tables(num_questions):
    """Returns a good number of tables for the given number of questions."""
    num_tables = 1
    for numq in PARAM_TABLE_LMITS:
        if numq >= num_questions:
            break
        else:
            num_tables += 1
    return num_tables


def _horizontal_line(row_geometry, num_choices, compact):
    parts = []
    num_empty_columns = 1 if not compact else 0
    first = 2
    extra_line = max(row_geometry) > 0 or -1 in row_geometry
    for geometry in row_geometry:
        if geometry > 0 or geometry == -1 or extra_line:
            parts.append(r"\cline{%d-%d}" % (first, first + num_choices - 1))
        first += 1 + num_empty_columns + num_choices
    return " ".join(parts)


def _table_top(num_tables, num_choices, compact, table_width=None, table_height=None):
    middle_sep_format = "p{3mm}" if not compact else ""
    middle_sep_header = " & & " if not compact else " & "
    lines = [r"\begin{center}", r"\large"]
    if table_width is not None and table_height is None:
        lines.append(r"\resizebox{%fcm}{!}{" % table_width)
    elif table_width is None and table_height is not None:
        lines.append(r"\resizebox{!}{%fcm}{" % table_height)
    elif table_width is not None and table_height is not None:
        lines.append(r"\resizebox{%fcm}{%fcm}{" % (table_width, table_height))
    column_flags = middle_sep_format.join(
        num_tables * ["|".join(["r"] + num_choices * ["c"] + [""])]
    )
    lines.append(r"\begin{tabular}{" + column_flags + "}")
    parts = []
    for __ in range(num_tables):
        parts_internal = []
        parts_internal.append(r"\multicolumn{1}{c}{}")
        for j in range(num_choices):
            parts_internal.append(r"\multicolumn{1}{c}{%s}" % chr(65 + j))
        parts.append(" & ".join(parts_internal))
    lines.append(middle_sep_header.join(parts) + r" \\")
    return lines


def _build_row(
    num_row,
    row_geometry,
    question_numbers,
    num_choices,
    infobits_row,
    compact,
    left_to_right_numbering=False,
):
    parts = []
    for i, geometry in enumerate(row_geometry):
        if geometry > 0:
            if not left_to_right_numbering:
                cell_number = num_row + question_numbers[i]
            else:
                cell_number = i + 1 + len(row_geometry) * num_row
            parts.append(_build_question_cell(cell_number, geometry))
        elif geometry == -1:
            parts.append(infobits_row[0][i])
        elif geometry == -2:
            parts.append(infobits_row[1][i])
        else:
            parts.append(r"\multicolumn{%d}{c}{}" % (1 + num_choices))
    row = " & & ".join(parts) if not compact else " & ".join(parts)
    return row + r" \\"


def _build_question_cell(num_question, num_choices):
    parts = [str(num_question)]
    for i in range(0, num_choices):
        parts.append(r"\light{%s}" % chr(65 + i))
    return " & ".join(parts)


def _create_infobits(bits, num_tables, num_choices, survey_mode):
    column_active = r"\multicolumn{1}{c}{$\blacksquare$}"
    column_inactive = r"\multicolumn{1}{c}{}"
    parts = [[], []]
    for i in range(0, num_tables):
        data = bits[i * num_choices : (i + 1) * num_choices]
        for j in (0, 1):
            val = j == 1
            components = [column_inactive]
            for bit in data:
                if not survey_mode and val ^ bit:
                    components.append(column_active)
                else:
                    components.append(column_inactive)
            parts[j].append(" & ".join(components))
    return parts


def format_questions(exam, model, with_solution=False):
    """Returns the questions of 'exam' formatted in LaTeX, as a string.

    'exam' is a create.questions.ExamQuestions object. It writes the questions
    in their 'shuffled' order. If 'with_solution', correct answers
    are marked in the text.

    """
    data = []
    if exam.num_questions() > 0:
        if model == "0":
            groups = exam.questions.groups
        else:
            groups = exam.shuffled_groups[model]
        counter = 0
        for group in groups:
            data.extend(
                format_group(group, model, counter, with_solution=with_solution)
            )
            counter += len(group)
        if model != "0":
            data.append("\n\n% solutions: ")
            solutions, _ = exam.solutions_and_permutations(model)
            data.append(" ".join([str(n) for n in solutions]))
            data.append("\n")
    return "".join(data)


def format_group(group, model, question_counter, with_solution=False):
    if model != "0":
        data = format_group_variation(
            group, model, question_counter, with_solution=with_solution
        )
    else:
        # Model 0: present all the variations
        data = []
        for variation in range(group.num_variations):
            group.select_variation(model, variation)
            if group.num_variations > 1:
                data.extend(
                    f"\\vspace{{0.8cm}}\\noindent\\emph{{[var. {variation + 1}]}}\n"
                )
            data.extend(
                format_group_variation(
                    group, model, question_counter, with_solution=with_solution
                )
            )
    return data


def format_group_variation(group, model, question_counter, with_solution=False):
    data = []
    if group.common_text is not None:
        common_component = group.common_text.component(model)
    else:
        common_component = None
    if len(group) > 1:
        data.append("\n\\vspace{0.1cm}\n")
        data.append("\\[ \\overbrace{\\hspace{0.8\\textwidth}} \\]\n")
        if common_component is None:
            data.append("\n\\vspace{-0.5cm}\n")
    if common_component is not None:
        right_block = _right_block_selected(common_component)
        if right_block:
            data.append(_start_right_block(common_component))
        data.extend(format_question_component(common_component))
        if right_block:
            data.extend(_end_right_block(common_component))
    data.append("\\begin{enumerate}[1.-]\n")
    data.append("\\setcounter{enumi}{" + str(question_counter) + "}")
    for question in group:
        data.append("\\vspace{2mm}\n")
        data.extend(format_question(question, model, with_solution))
        data.append("\n")
    data.append("\\end{enumerate}\n")
    if len(group) > 1:
        data.append("\n\\vspace{-0.8cm}\n")
        data.append("\\[ \\underbrace{\\hspace{0.8\\textwidth}} \\]\n")
        data.append("\n\\vspace{0.1cm}\n")
    return data


def format_question(question, model, with_solution=False):
    """Returns a latex formatted question, as a list of strings.

    If 'with_solution', correct answers are marked in the text.

    """
    data = []
    choices = question.shuffled_choices(model)
    text_component = question.text(model)
    right_block = _right_block_selected(text_component)
    if right_block:
        data.append(_start_right_block(text_component))
    data.append(r"\item ")
    data.extend(format_question_component(text_component))
    data.append("\n  \\begin{enumerate}[(a)]\n")
    for choice in choices:
        data.append(r"    \item ")
        if with_solution and choice in question.correct_choices(model):
            data[-1] = data[-1] + r" \textbf{***} "
        data.extend(format_question_component(choice))
        data.append("\n")
    data.append("\n  \\end{enumerate}\n")
    if right_block:
        data.extend(_end_right_block(text_component))
    return data


def _right_block_selected(question_component):
    return (
        question_component.figure is not None or question_component.code is not None
    ) and question_component.annex_pos == "right"


def _start_right_block(question_component):
    width_right = question_component.annex_width + PARAM_TABLE_SEP
    width_left = 1 - width_right - PARAM_TABLE_MARGIN
    if question_component.annex_separation:
        separation = "|"
    else:
        separation = ""
    return (
        "\\hspace{-0.2cm}"
        f"\\begin{{tabular}}[l]{{p{{{width_left}\\textwidth}}"
        f"{separation}"
        f"p{{{width_right}\\textwidth}}}}\n"
    )


def _end_right_block(question_component):
    data = ["&\n"]
    if question_component.figure is not None:
        data.extend(
            write_figure(
                question_component.figure, question_component.annex_width, True
            )
        )
    elif question_component.code is not None:
        data.extend(write_code(question_component.code))
    data.append("\\\\\n\\end{tabular}\n")
    return data


def format_question_component(component):
    data = []
    if component.text is not None:
        if isinstance(component.text, str):
            data.append(component.text)
        else:
            for part in component.text:
                if part[0] == "text":
                    data.append(part[1])
                elif part[0] == "code":
                    data.extend(write_code(part[1]))
    if component.figure is not None and component.annex_pos == "center":
        data.extend(write_figure(component.figure, component.annex_width, True))
    elif component.code is not None and component.annex_pos == "center":
        data.extend(write_code(component.code))
    return data


def write_figure(figure, width, center):
    data = []
    if center:
        data.append("\\begin{center}\n")
    data.append("\\includegraphics[width=%f\\textwidth]{%s}\n" % (width * 0.9, figure))
    if center:
        data.append("\\end{center}\n")
    return data


def write_code(code):
    data = []
    data.append("\\begin{center}\n" "\\begin{verbatim}\n")
    data.append(code + "\n")
    data.append("\\end{verbatim}\n" "\\end{center}")
    return data


def re_id_box_replacer(match):
    """Takes a re.match object and returns the id box.

    Two groups expected: (1) number of digits; (2) label to show.

    """
    return create_id_box(match.group(2), int(match.group(1)))



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2021 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <https://www.gnu.org/licenses/>.
#

from PyQt6.QtGui import QIcon, QImage

from PyQt6.QtWidgets import (
    QListView,
    QListWidget,
    QListWidgetItem,
)

from PyQt6.QtCore import QEvent, QItemSelection, QObject, QSize, pyqtSignal, pyqtSlot

from .. import exams


class ExamIcon(QIcon):
    def __init__(self, exam):
        super().__init__(exam.image_drawn_path())


class ExamImage(QImage):
    def __init__(self, exam):
        super().__init__(exam.image_drawn_path())


class ThumbnailsViewItem(QListWidgetItem):
    def __init__(self, exam):
        self.exam = exam
        super().__init__(ExamIcon(exam), self._label())

    def update(self):
        self.setText(self._label())
        self.setIcon(ExamIcon(self.exam))

    def _label(self):
        if self.exam.decisions.student:
            label = self.exam.decisions.student.name_or_id
        else:
            label = ""
        return label


class ThumbnailsView(QListWidget):
    selection_changed = pyqtSignal(exams.Exam)

    def __init__(self, parent):
        super().__init__(parent)
        self.setIconSize(QSize(120, 80))
        self.setViewMode(QListView.ViewMode.IconMode)
        self.setMovement(QListView.Movement.Static)
        self.setUniformItemSizes(True)
        self.setResizeMode(QListView.ResizeMode.Adjust)
        self.exams = []
        self.selectionModel().selectionChanged.connect(self.on_selection)
        self.keyboard_filter = KeyboardEventsFilter()
        self.installEventFilter(self.keyboard_filter)

    def add_exams(self, exams):
        for exam in exams:
            self.add_exam(exam, scroll=False)

    def add_exam(self, exam, scroll=True):
        self.addItem(ThumbnailsViewItem(exam))
        if scroll:
            self.scrollToBottom()
        self.exams.append(exam)

    def clear_exams(self):
        self.clear()
        self.exams = []

    def update_exam(self, exam):
        pos = self.exams.index(exam)
        self.item(pos).update()

    def remove_exam(self, exam):
        pos = self.exams.index(exam)
        del self.exams[pos]
        self.takeItem(pos)

    def selected_exam(self):
        items = self.selectedItems()
        if items:
            return items[0].exam
        else:
            return None

    def select_next_exam(self):
        current_exam = self.selected_exam()
        if current_exam is not None:
            pos = 1 + self.exams.index(current_exam)
            if pos < len(self.exams):
                item = self.item(pos)
                item.setSelected(True)
                self.scrollToItem(item)

    def clear_selected_exam(self):
        items = self.selectedItems()
        if items:
            items[0].setSelected(False)

    def block_keyboard(self, block):
        self.keyboard_filter.setBlocking(block)

    @pyqtSlot(QItemSelection, QItemSelection)
    def on_selection(self, selected, deselected):
        indexes = selected.indexes()
        if len(indexes):
            selected = indexes[0].row()
            if selected < len(self.exams):
                self.selection_changed.emit(self.exams[selected])


class KeyboardEventsFilter(QObject):
    def __init__(self):
        self.blocking = True
        super().__init__()

    def setBlocking(self, blocking):
        self.blocking = blocking

    def eventFilter(self, widget, event):
        if self.blocking and event.type() == QEvent.Type.KeyPress:
            return True
        else:
            return super().eventFilter(widget, event)



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2024 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <https://www.gnu.org/licenses/>.
#


from dataclasses import dataclass
import enum
from typing import Optional, Union, Sequence
import statistics

from .. import exams


@dataclass(frozen=True)
class QuestionPermutations:
    reference_model: str
    question_permutations: dict[str, list[int]]

    def unwind_question(self, question: int, model: str) -> int:
        # Input question numbers are 1-based
        # Returned question numbers are 0-based
        if model == self.reference_model:
            return question - 1
        elif model in self.question_permutations:
            return self.question_permutations[model][question - 1] - 1
        else:
            raise ValueError(f"Model {model} not defined in permutations")


@dataclass(frozen=True)
class ChoicePermutations:
    num_choices: int
    reference_model: str
    choice_permutations: dict[str, list[int]]

    def __post_init__(self):
        if not len(self.choice_permutations):
            raise ValueError("No permutations have been defined")

    def unwind_choice(self, choice: int, model: str) -> int:
        if choice == 0:
            # Blank answer
            return 0
        elif model == self.reference_model:
            return choice
        elif model in self.choice_permutations:
            return self.choice_permutations[model][choice - 1]
        else:
            raise ValueError(f"Model {model} not defined in permutations")

    def reference_choices(self, model: str) -> list[int]:
        if model == self.reference_model:
            return list(range(1, self.num_choices + 1))
        elif model in self.choice_permutations:
            return self.choice_permutations[model]
        else:
            raise ValueError(f"Model {model} not defined in permutations")


@dataclass(frozen=True)
class ExamPermutations:
    num_questions: int
    num_choices: int
    reference_model: str
    exam_permutations: dict[str, list[tuple[int, list[int]]]]

    def get_question_permutations(self) -> QuestionPermutations:
        return QuestionPermutations(
            reference_model=self.reference_model,
            question_permutations={
                model: [question for question, _ in self.exam_permutations[model]]
                for model in self.exam_permutations
            },
        )

    def unwind_choice_permutations(self) -> list[ChoicePermutations]:
        choice_permutations_list: list[dict[str, list[int]]] = [
            {} for _ in range(self.num_questions)
        ]
        for model, p in self.exam_permutations.items():
            for question_num, choice_permutations in p:
                choice_permutations_list[question_num - 1][model] = choice_permutations
        return [
            ChoicePermutations(
                num_choices=self.num_choices,
                reference_model=self.reference_model,
                choice_permutations=choice_permutations,
            )
            for choice_permutations in choice_permutations_list
        ]


class StatsType(enum.Enum):
    CORRECT = enum.auto()
    INCORRECT = enum.auto()
    BLANK = enum.auto()
    SCORE = enum.auto()


@dataclass
class AbstractStats:
    stats_type: StatsType
    num_exams: int
    average: float
    median: Union[float, int]
    min: Union[float, int]
    max: Union[float, int]


@dataclass
class Stats:
    overall: dict[StatsType, Optional[AbstractStats]]
    q_by_q: "QByQStats"


class OverallStats:
    correct_answers: list[int]
    wrong_answers: list[int]
    blank_answers: list[int]
    scores: list[float]

    def __init__(self) -> None:
        self.correct_answers = []
        self.wrong_answers = []
        self.blank_answers = []
        self.scores = []

    def count_answers(self, exam: exams.Exam) -> None:
        self.correct_answers.append(exam.score.correct)
        self.wrong_answers.append(exam.score.incorrect)
        self.blank_answers.append(exam.score.blank)
        if exam.score.score is not None:
            self.scores.append(exam.score.score)

    def get_stats(self, stats_type: StatsType) -> Optional[AbstractStats]:
        data: Sequence[Union[float, int]]
        if stats_type == StatsType.CORRECT:
            data = self.correct_answers
        elif stats_type == StatsType.INCORRECT:
            data = self.wrong_answers
        elif stats_type == StatsType.BLANK:
            data = self.blank_answers
        elif stats_type == StatsType.SCORE:
            data = self.scores
        else:
            raise ValueError(f"Invalid stats type: {stats_type}")
        if data:
            return self._compute_stats(data, stats_type)
        else:
            return None

    def get_all_stats(self) -> dict[StatsType, Optional[AbstractStats]]:
        return {stats_type: self.get_stats(stats_type) for stats_type in StatsType}

    def _compute_stats(
        self, data: Sequence[Union[float, int]], stats_type: StatsType
    ) -> AbstractStats:
        if not data:
            raise ValueError("No data to compute stats")
        return AbstractStats(
            stats_type=stats_type,
            num_exams=len(data),
            average=statistics.fmean(data),
            median=statistics.median_high(data),
            min=min(data),
            max=max(data),
        )


class QByQStats:
    question_permutations: Optional[QuestionPermutations]
    question_stats: list["QuestionStats"]
    num_questions: int
    num_choices: int

    def __init__(
        self,
        num_questions: int,
        num_choices: int,
        correct_choices_dict: dict[str, list[set[int]]],
        permutations: Optional[ExamPermutations],
    ) -> None:
        self.num_questions = num_questions
        self.num_choices = num_choices
        self.question_stats = []
        if permutations is not None:
            self.question_permutations = permutations.get_question_permutations()
            choice_permutations_list = permutations.unwind_choice_permutations()
            correct_choices_dict = self._correct_choices_for_reference_model(
                correct_choices_dict, choice_permutations_list
            )
            correct_choices_list = self._unwind_correct_choices(
                num_questions, correct_choices_dict
            )
            self.question_stats = [
                QuestionStats(num_choices, correct_choices, permutations)
                for correct_choices, permutations in zip(
                    correct_choices_list, choice_permutations_list
                )
            ]
        else:
            self.question_permutations = None
            correct_choices_list = self._unwind_correct_choices(
                num_questions, correct_choices_dict
            )
            self.question_stats = [
                QuestionStats(num_choices, correct_choices, None)
                for correct_choices in correct_choices_list
            ]

    @property
    def models(self) -> list[str]:
        if self.question_permutations is not None:
            return sorted(
                [key for key in self.question_permutations.question_permutations]
            )
        elif self.question_stats:
            return self.question_stats[0].models
        else:
            raise ValueError("No models have been defined")

    @property
    def reference_model(self) -> Optional[str]:
        if self.question_permutations is not None:
            return self.question_permutations.reference_model
        else:
            return None

    @property
    def all_models(self) -> list[str]:
        models = self.models
        if (
            self.question_permutations is not None
            and self.question_permutations.reference_model not in models
        ):
            return [self.question_permutations.reference_model] + models
        else:
            return models

    def count_answer(self, answer: int, question: int, model: str) -> None:
        # Question numbers are 1-based
        if self.question_permutations is not None:
            reference_question = self.question_permutations.unwind_question(
                question, model
            )
        else:
            reference_question = question - 1
        self.question_stats[reference_question].count_answer(answer, model)

    def count_answers(self, answers: list[int], model: str) -> None:
        for question, answer in enumerate(answers):
            self.count_answer(answer, question + 1, model)

    def get_answer_counts(self, model: str) -> list[list[int]]:
        return [q.get_answer_counts(model) for q in self._reorder_question_stats(model)]

    def get_answer_ratios(self, model: str) -> list[list[float]]:
        return [q.get_answer_ratios(model) for q in self._reorder_question_stats(model)]

    def get_answer_percentages(self, model: str) -> list[list[float]]:
        return [
            q.get_answer_percentages(model) for q in self._reorder_question_stats(model)
        ]

    def is_correct(self, answer: int, question: int, model: str) -> bool:
        # Question numbers are 1-based
        if self.question_permutations is not None:
            reference_question = self.question_permutations.unwind_question(
                question, model
            )
        else:
            reference_question = question - 1
        return self.question_stats[reference_question].is_correct(answer, model)

    def _reorder_question_stats(self, model: str) -> list["QuestionStats"]:
        if self.question_permutations is not None and model != self.reference_model:
            return [
                self.question_stats[q - 1]
                for q in self.question_permutations.question_permutations[model]
            ]
        else:
            return list(self.question_stats)

    @staticmethod
    def _unwind_correct_choices(
        num_questions: int, correct_choices: dict[str, list[set[int]]]
    ) -> list[dict[str, set[int]]]:
        return [
            {model: correct_choices[model][i] for model in correct_choices.keys()}
            for i in range(num_questions)
        ]

    def _correct_choices_for_reference_model(
        self,
        correct_choices_dict: dict[str, list[set[int]]],
        choice_permutations_list: list[ChoicePermutations],
    ) -> dict[str, list[set[int]]]:
        if self.question_permutations is None:
            raise ValueError("No question permutations have been defined")
        if not correct_choices_dict:
            raise ValueError("No correct choices have been defined for any model")
        reference_model = self.question_permutations.reference_model
        if reference_model in correct_choices_dict:
            correct_choices = {reference_model: correct_choices_dict[reference_model]}
        else:
            model = next(iter(correct_choices_dict))
            correct_choices = {
                reference_model: [
                    set() for _ in range(len(correct_choices_dict[model]))
                ]
            }
            for i, choices in enumerate(correct_choices_dict[model]):
                question = self.question_permutations.unwind_question(i + 1, model)
                correct_choices[reference_model][question] = {
                    choice_permutations_list[question].unwind_choice(choice, model)
                    for choice in choices
                }
        return correct_choices


class QuestionStats:
    correct_choices: dict[str, set[int]]
    permutations: Optional[ChoicePermutations]
    answer_counts: dict[str, list[int]]

    def __init__(
        self,
        num_choices,
        correct_choices: dict[str, set[int]],
        permutations: Optional[ChoicePermutations],
    ) -> None:
        self.correct_choices = correct_choices
        self.permutations = permutations
        self.answer_counts = {
            model: [0] * (num_choices + 1)  # pos. 0 is for blank answers
            for model in correct_choices
        }

    @property
    def models(self) -> list[str]:
        return sorted([key for key in self.answer_counts])

    def count_answer(self, answer: int, model: str) -> None:
        if self.permutations is not None:
            reference_choice = self.permutations.unwind_choice(answer, model)
            self.answer_counts[self.permutations.reference_model][reference_choice] += 1
        else:
            self.answer_counts[model][answer] += 1

    def get_answer_counts(self, model: str) -> list[int]:
        if self.permutations is not None:
            reference_choices = [0] + self.permutations.reference_choices(model)
            return [
                self.answer_counts[self.permutations.reference_model][choice]
                for choice in reference_choices
            ]
        else:
            return list(self.answer_counts[model])

    def get_answer_ratios(self, model: str) -> list[float]:
        counts = self.get_answer_counts(model)
        total = sum(counts)
        if total > 0:
            return [count / total for count in counts]
        else:
            return [0.0] * len(counts)

    def get_answer_percentages(self, model: str) -> list[float]:
        return [ratio * 100 for ratio in self.get_answer_ratios(model)]

    def is_correct(self, answer: int, model: str) -> bool:
        if self.permutations is not None:
            reference_choice = self.permutations.unwind_choice(answer, model)
            return (
                reference_choice
                in self.correct_choices[self.permutations.reference_model]
            )
        elif model in self.correct_choices:
            return answer in self.correct_choices[model]
        else:
            raise ValueError(f"Model {model} not defined in correct choices")



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2024 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <https://www.gnu.org/licenses/>.
#

from . import compute_stats

import argparse


def print_answer_counts(answer_counts: list[list[int]]) -> None:
    for i, counts in enumerate(answer_counts):
        print(f"[Q{i + 1}] {_format_choice_counts(counts)}")


def _format_choice_counts(choice_counts: list[int]) -> str:
    not_blank = ", ".join(
        f"{chr(65 + i)}: {count}" for i, count in enumerate(choice_counts[1:])
    )
    blank = f", blank: {choice_counts[0]}"
    return not_blank + blank


def _cmd_options():
    parser = argparse.ArgumentParser(
        description="Compute statistics from a session database"
    )
    parser.add_argument(
        "session",
        type=str,
        help="Path to the session database",
    )
    return parser.parse_args()


def main():
    args = _cmd_options()
    stats = compute_stats.stats_from_session_path(args.session)
    reference_model = stats.reference_model
    if reference_model is not None:
        print_answer_counts(stats.get_answer_counts(reference_model))
    else:
        for model in stats.models:
            print(f"Model {model}")
            print_answer_counts(stats.get_answer_counts(model))


if __name__ == "__main__":
    main()



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2021 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#

import unittest
import configparser
import fractions

import eyegrade.exams as exams
import eyegrade.scoring as scoring
import eyegrade.utils as utils


class TestExamConfigParser(unittest.TestCase):
    def test_correct_file(self):
        config_text = """
            [exam]
                dimensions: 3,5
                id-num-digits: 9
                correct-weight: 1
                incorrect-weight: 1/2
                blank-weight: 0

                [solutions]
                model-A: 3/1/2/1/2
                model-B: 1/3/2/2/3
        """
        exam_data = configparser.ConfigParser()
        exam_data.read_string(config_text)
        exam_config = exams.ExamConfig()
        exam_config._read_config_parser(exam_data)
        self.assertEqual(exam_config.solutions["A"], [{3}, {1}, {2}, {1}, {2}])
        self.assertEqual(exam_config.solutions["B"], [{1}, {3}, {2}, {2}, {3}])
        self.assertEqual(exam_config.id_num_digits, 9)
        score = scoring.QuestionScores(1, fractions.Fraction(1, 2), 0)
        for model in ["A", "B"]:
            for i in range(5):
                self.assertEqual(exam_config.scores[model][i], score)
        self.assertEqual(exam_config.dimensions, [(3, 5)])

    def test_solutions_out_of_range(self):
        config_text = """
            [exam]
                dimensions: 3,5
                id-num-digits: 9
                correct-weight: 1
                incorrect-weight: 1/2
                blank-weight: 0

                [solutions]
                model-A: 3/1/4/1/2
                model-B: 1/3/2/2/3
        """
        exam_data = configparser.ConfigParser()
        exam_data.read_string(config_text)
        exam_config = exams.ExamConfig()
        with self.assertRaises(utils.EyegradeException):
            exam_config._read_config_parser(exam_data)



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2024 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <https://www.gnu.org/licenses/>.

import gettext
from typing import Optional, Union

from PyQt6.QtWidgets import (
    QDialog,
    QVBoxLayout,
    QDialogButtonBox,
    QTabWidget,
    QTableView,
    QSizePolicy,
    QWidget,
)

from PyQt6.QtCore import QAbstractTableModel, QModelIndex, Qt, QVariant

from PyQt6.QtGui import QBrush, QColor

from . import widgets
from ..stats import stats_core
from .. import utils


t = gettext.translation("eyegrade", utils.locale_dir(), fallback=True)
_ = t.gettext


class DialogShowStats(QDialog):
    """Dialog to show question by question statistics.

    Example (replace `parent` by the parent widget):

    dialog = DialogShowStats(parent)
    dialog.exec()

    """

    def __init__(self, parent: QWidget, stats: stats_core.Stats) -> None:
        super().__init__(parent)
        self.setWindowTitle(_("Exam statistics"))
        layout = QVBoxLayout(self)
        self.setLayout(layout)
        buttons = QDialogButtonBox(QDialogButtonBox.StandardButton.Ok)
        buttons.accepted.connect(self.accept)
        tabs = QTabWidget(self)
        overall_stats_widget = OverallStatsWidget(self, stats.overall)
        q_by_q_stats_widget = QByQStatsWidget(self, stats.q_by_q)
        tabs.addTab(overall_stats_widget, _("Overall"))
        tabs.addTab(q_by_q_stats_widget, _("Question by question"))
        layout.addWidget(tabs)
        layout.addWidget(buttons)
        self.adjustSize()
        self.setModal(True)


class OverallStatsWidget(QWidget):
    overall_stats: dict[stats_core.StatsType, Optional[stats_core.AbstractStats]]

    def __init__(
        self,
        parent: QWidget,
        overall_stats: dict[stats_core.StatsType, Optional[stats_core.AbstractStats]],
    ) -> None:
        super().__init__(parent)
        self.overall_stats = overall_stats
        self.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum)
        layout = QVBoxLayout(self)
        self.setLayout(layout)
        table = self._create_table(self)
        layout.addWidget(table)
        layout.setAlignment(table, Qt.AlignmentFlag.AlignHCenter)

    def _create_table(self, parent) -> QTableView:
        table = widgets.CustomTableView(
            parent=parent, maximum_width=750, maximum_height=500
        )
        table.setModel(OverallStatsTableModel(self.overall_stats))
        table.adjust_size()
        return table


class QByQStatsWidget(QTabWidget):
    q_by_q_stats: stats_core.QByQStats

    def __init__(self, parent: QWidget, q_by_q_stats: stats_core.QByQStats) -> None:
        super().__init__(parent)
        self.q_by_q_stats = q_by_q_stats
        self.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum)
        for model in self.q_by_q_stats.all_models:
            widget = QWidget(self)
            tab_layout = QVBoxLayout(widget)
            widget.setLayout(tab_layout)
            table = self._create_table(model, widget)
            tab_layout.addWidget(table, Qt.AlignmentFlag.AlignCenter)
            tab_layout.setAlignment(table, Qt.AlignmentFlag.AlignHCenter)
            self.addTab(widget, _("Model ") + model)

    def _create_table(self, model: str, parent) -> QTableView:
        table = widgets.CustomTableView(
            parent=parent, maximum_width=750, maximum_height=500
        )
        table.setModel(QByQStatsTableModel(self.q_by_q_stats, model))
        table.adjust_size()
        return table


class OverallStatsTableModel(QAbstractTableModel):
    overall_stats: dict[stats_core.StatsType, Optional[stats_core.AbstractStats]]
    row_headers = [
        _("Correct answers"),
        _("Incorrect answers"),
        _("Blank answers"),
        _("Score"),
    ]
    column_headers = [_("# exams"), _("Average"), _("Median"), _("Min."), _("Max.")]
    row_types = [
        stats_core.StatsType.CORRECT,
        stats_core.StatsType.INCORRECT,
        stats_core.StatsType.BLANK,
        stats_core.StatsType.SCORE,
    ]

    def __init__(
        self,
        overall_stats: dict[stats_core.StatsType, Optional[stats_core.AbstractStats]],
        parent: Optional[QWidget] = None,
    ) -> None:
        super().__init__(parent=parent)
        self.overall_stats = overall_stats

    def rowCount(self, parent=QModelIndex()) -> int:
        return 4

    def columnCount(self, parent=QModelIndex()) -> int:
        return 5

    def headerData(
        self,
        index: int,
        orientation: Qt.Orientation,
        role: int = Qt.ItemDataRole.DisplayRole,
    ) -> Union[str, QVariant]:
        if role == Qt.ItemDataRole.DisplayRole:
            if orientation == Qt.Orientation.Horizontal:
                headers = self.column_headers
            else:
                headers = self.row_headers
            return headers[index]
        else:
            return QVariant()

    def data(
        self, index: QModelIndex, role: int = Qt.ItemDataRole.DisplayRole
    ) -> Union[str, Qt.AlignmentFlag, QVariant]:
        if role == Qt.ItemDataRole.DisplayRole:
            r = index.row()
            c = index.column()
            stats_data = self.overall_stats[self.row_types[r]]
            if stats_data is not None:
                if c == 0:
                    return str(stats_data.num_exams)
                elif c == 1:
                    return self._fmt(stats_data.average)
                elif c == 2:
                    return self._fmt(stats_data.median)
                elif c == 3:
                    return self._fmt(stats_data.min)
                elif c == 4:
                    return self._fmt(stats_data.max)
                else:
                    return QVariant()
            else:
                return "-"
        elif role == Qt.ItemDataRole.TextAlignmentRole:
            return Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter
        else:
            return QVariant()

    def flags(self, index: QModelIndex) -> Qt.ItemFlag:
        return Qt.ItemFlag(Qt.ItemFlag.ItemIsEnabled)

    def _fmt(self, value: Union[int, float]) -> str:
        if isinstance(value, int):
            return str(value)
        else:
            return f"{value:.2f}"


class QByQStatsTableModel(QAbstractTableModel):
    q_by_q_stats: stats_core.QByQStats
    answer_counts: list[list[int]]
    answer_percentages: list[list[float]]
    model: str

    def __init__(
        self,
        q_by_q_stats: stats_core.QByQStats,
        model: str,
        parent: Optional[QWidget] = None,
    ) -> None:
        super().__init__(parent=parent)
        self.q_by_q_stats = q_by_q_stats
        self.answer_counts = self.q_by_q_stats.get_answer_counts(model)
        self.answer_percentages = self.q_by_q_stats.get_answer_percentages(model)
        self.model = model

    def rowCount(self, parent=QModelIndex()) -> int:
        return self.q_by_q_stats.num_questions

    def columnCount(self, parent=QModelIndex()) -> int:
        return self.q_by_q_stats.num_choices + 1

    def headerData(
        self,
        index: int,
        orientation: Qt.Orientation,
        role: int = Qt.ItemDataRole.DisplayRole,
    ) -> Union[str, QVariant]:
        if role == Qt.ItemDataRole.DisplayRole:
            if orientation == Qt.Orientation.Horizontal:
                if index < self.q_by_q_stats.num_choices:
                    return "{} {}".format(_("Answer"), chr(65 + index))
                else:
                    return _("Blank")
            else:
                return "{0} {1}".format(_("Question"), index + 1)
        else:
            return QVariant()

    def data(
        self, index: QModelIndex, role: int = Qt.ItemDataRole.DisplayRole
    ) -> Union[str, Qt.AlignmentFlag, QVariant]:
        if role == Qt.ItemDataRole.DisplayRole:
            r = index.row()
            c = index.column()
            if c < self.q_by_q_stats.num_choices:
                return str(self.answer_counts[r][c + 1])
            else:
                return str(self.answer_counts[r][0])
        elif role == Qt.ItemDataRole.ToolTipRole:
            r = index.row()
            c = index.column()
            if c < self.q_by_q_stats.num_choices:
                return f"{self.answer_percentages[r][c + 1]:.1f}%"
            else:
                return f"{self.answer_percentages[r][0]:.1f}%"
        elif role == Qt.ItemDataRole.TextAlignmentRole:
            return Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter
        elif role == Qt.ItemDataRole.BackgroundRole:
            r = index.row()
            c = index.column()
            if c < self.q_by_q_stats.num_choices and self.q_by_q_stats.is_correct(
                c + 1, r + 1, self.model
            ):
                return QVariant(QBrush(QColor(205, 255, 212)))
            else:
                return QVariant()
        else:
            return QVariant()

    def flags(self, index: QModelIndex) -> Qt.ItemFlag:
        return Qt.ItemFlag(Qt.ItemFlag.ItemIsEnabled)
